import ast
from typing import List, Optional

from pydantic_settings import SettingsConfigDict
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.language.ast import OperationDefinitionNode, OperationType
from turms.recurse import type_field_node
from turms.plugins.base import Plugin, PluginConfig
from pydantic import Field
from graphql.language.ast import (
    FieldNode,
    OperationDefinitionNode,
    OperationType,
)
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.utilities.get_operation_root_type import get_operation_root_type
from graphql.utilities.type_info import get_field_def

import re
from graphql import NonNullTypeNode, VariableDefinitionNode, language
from turms.registry import ClassRegistry
from turms.utils import (
    generate_pydantic_config,
    inspect_operation_for_documentation,
    parse_documents,
    recurse_type_annotation,
    replace_iteratively,
    parse_value_node,
)
import logging


logger = logging.getLogger(__name__)
fragment_searcher = re.compile(r"\.\.\.(?P<fragment>[a-zA-Z]*)")


class OperationsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(env_prefix="TURMS_PLUGINS_OPERATIONS_")
    type: str = "turms.plugins.operations.OperationsPlugin"
    query_bases: List[str] = []
    arguments_bases: List[str] = []
    mutation_bases: List[str] = []
    subscription_bases: List[str] = []
    operations_glob: Optional[str] = None
    create_arguments: bool = True
    extract_documentation: bool = True
    arguments_allow_population_by_field_name: bool = False


def get_query_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    if plugin_config.query_bases:
        for base in plugin_config.query_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.query_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

            return [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.object_bases
            ]


def get_mutation_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    if plugin_config.mutation_bases:
        for base in plugin_config.mutation_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.mutation_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

            return [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.object_bases
            ]


def generate_arguments_config(
    operation: OperationDefinitionNode,
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):

    if config.pydantic_version == "1":
        config_fields = []

        if plugin_config.arguments_allow_population_by_field_name:
            config_fields.append(
                ast.Assign(
                    targets=[
                        ast.Name(id="allow_population_by_field_name", ctx=ast.Store())
                    ],
                    value=ast.Constant(value=True),
                )
            )

        if len(config_fields) > 0:
            return [
                ast.ClassDef(
                    name="Config",
                    bases=[],
                    keywords=[],
                    body=config_fields,
                    decorator_list=[],
                )
            ]
        else:
            return []

    else:

        config_keywords = []

        if plugin_config.arguments_allow_population_by_field_name is not None:
            config_keywords.append(
                ast.keyword(
                    arg="populate_by_name",
                    value=ast.Constant(
                        value=config.options.allow_population_by_field_name
                    ),
                )
            )

        if len(config_keywords) > 0:
            registry.register_import("pydantic.ConfigDict")
            return [
                ast.Assign(
                    targets=[ast.Name(id="model_config", ctx=ast.Store())],
                    value=ast.Call(
                        func=ast.Name(id="ConfigDict", ctx=ast.Load()),
                        args=[],
                        keywords=config_keywords,
                    ),
                )
            ]
        else:
            return []


def get_arguments_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    if plugin_config.arguments_bases:
        for base in plugin_config.arguments_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.arguments_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

            return [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.object_bases
            ]


def get_subscription_bases(
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    if plugin_config.subscription_bases:
        for base in plugin_config.subscription_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in plugin_config.subscription_bases
        ]
    else:
        for base in config.object_bases:
            registry.register_import(base)

            return [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.object_bases
            ]


def generate_expanded_types(
    v: VariableDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):

    if isinstance(v.type, NonNullTypeNode):
        type_node = v.type.type
    else:
        type_node = v.type

    type_name = type_node.name.value
    type_definition = client_schema.get_type(type_name)

    if type_definition is None:
        return []

    if type_definition.ast_node is None:
        return []

    if type_definition.ast_node.kind != "InputObjectTypeDefinition":
        return []

    fields = type_definition.ast_node.fields

    fields_body = []

    additional_keywords = []

    for field in fields:
        field_name = field.name.value
        field_type = field.type

        is_optional = not isinstance(field_type, NonNullTypeNode)
        annotation = recurse_type_annotation(field_type, registry)

        if is_optional:
            assign = ast.AnnAssign(
                target=ast.Name(field_name, ctx=ast.Store()),
                annotation=annotation,
                value=ast.Call(
                    func=ast.Name(id="Field", ctx=ast.Load()),
                    args=[],
                    keywords=[
                        ast.keyword(arg="alias", value=ast.Constant(value=field_name))
                    ],
                ),
                simple=1,
            )
        else:
            assign = ast.AnnAssign(
                target=ast.Name(field_name, ctx=ast.Store()),
                annotation=annotation,
                simple=1,
            )

        fields_body += [assign]


def generate_operation(
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: OperationsPluginConfig,
    registry: ClassRegistry,
):
    tree = []
    assert o.name.value, "Operation names are required"

    # Generation means creating a class for the operation
    if o.operation == OperationType.MUTATION:
        class_name = registry.generate_mutation(o.name.value)
        extra_bases = get_mutation_bases(config, plugin_config, registry)
    if o.operation == OperationType.SUBSCRIPTION:
        class_name = registry.generate_subscription(o.name.value)
        extra_bases = get_subscription_bases(config, plugin_config, registry)
    if o.operation == OperationType.QUERY:
        class_name = registry.generate_query(o.name.value)
        extra_bases = get_query_bases(config, plugin_config, registry)

    x = get_operation_root_type(client_schema, o)
    class_body_fields = []

    operation_documentation = (
        inspect_operation_for_documentation(o)
        if plugin_config.extract_documentation
        else None
    )
    if operation_documentation:
        class_body_fields.append(
            ast.Expr(
                value=ast.Constant(value=operation_documentation),
            )
        )

    for field_node in o.selection_set.selections:
        field_node: FieldNode = field_node
        field_definition = get_field_def(client_schema, x, field_node)
        assert field_definition, "Couldn't find field definition"

        class_body_fields += [
            type_field_node(
                field_node,
                class_name,
                field_definition,
                client_schema,
                config,
                tree,
                registry,
            ),
        ]

    query_document = language.print_ast(o)
    merged_document = replace_iteratively(query_document, registry)

    if plugin_config.create_arguments:
        arguments_body = []

        for v in o.variable_definitions:
            is_optional = not isinstance(v.type, NonNullTypeNode) or v.default_value
            annotation = recurse_type_annotation(v.type, registry)
            field_name = registry.generate_parameter_name(v.variable.name.value)
            target = v.variable.name.value

            if target != field_name:
                registry.register_import("pydantic.Field")
                if is_optional:
                    assign = ast.AnnAssign(
                        target=ast.Name(field_name, ctx=ast.Store()),
                        annotation=annotation,
                        value=ast.Call(
                            func=ast.Name(id="Field", ctx=ast.Load()),
                            args=[],
                            keywords=[
                                ast.keyword(
                                    arg="alias", value=ast.Constant(value=target)
                                ),
                                ast.keyword(
                                    arg="default",
                                    value=ast.Constant(
                                        value=(
                                            parse_value_node(v.default_value)
                                            if v.default_value is not None
                                            else None
                                        )
                                    ),
                                ),
                            ],
                        ),
                        simple=1,
                    )
                else:
                    assign = ast.AnnAssign(
                        target=ast.Name(field_name, ctx=ast.Store()),
                        annotation=annotation,
                        value=ast.Call(
                            func=ast.Name(id="Field", ctx=ast.Load()),
                            args=[],
                            keywords=[
                                ast.keyword(
                                    arg="alias", value=ast.Constant(value=target)
                                )
                            ],
                        ),
                        simple=1,
                    )
            else:
                if is_optional:
                    assign = ast.AnnAssign(
                        target=ast.Name(field_name, ctx=ast.Store()),
                        annotation=annotation,
                        value=ast.Call(
                            func=ast.Name(id="Field", ctx=ast.Load()),
                            args=[],
                            keywords=[
                                ast.keyword(
                                    arg="default",
                                    value=ast.Constant(
                                        value=(
                                            parse_value_node(v.default_value)
                                            if v.default_value is not None
                                            else None
                                        )
                                    ),
                                ),
                            ],
                        ),
                        simple=1,
                    )
                else:
                    assign = ast.AnnAssign(
                        target=ast.Name(field_name, ctx=ast.Store()),
                        annotation=annotation,
                        simple=1,
                    )

            arguments_body += [assign]

        arguments_body + generate_arguments_config(
            o.operation, config, plugin_config, registry
        )

        class_body_fields += [
            ast.ClassDef(
                "Arguments",
                bases=get_arguments_bases(config, plugin_config, registry=registry),
                decorator_list=[],
                keywords=[],
                body=arguments_body or [ast.Pass()],
            )
        ]

    meta_body = [
        ast.Assign(
            targets=[ast.Name(id="document", ctx=ast.Store())],
            value=ast.Constant(value=merged_document),
        ),
    ]
    if config.domain:
        meta_body += [
            ast.Assign(
                targets=[ast.Name(id="domain", ctx=ast.Store())],
                value=ast.Constant(value=str(config.domain)),
            )
        ]

    class_body_fields += [
        ast.ClassDef("Meta", bases=[], decorator_list=[], keywords=[], body=meta_body)
    ]

    tree.append(
        ast.ClassDef(
            class_name,
            bases=extra_bases,
            decorator_list=[],
            keywords=[],
            body=class_body_fields
            + generate_pydantic_config(o.operation, config, registry),
        )
    )

    return tree


class OperationsPlugin(Plugin):
    """ " Generate operations as classes

    This plugin created classes for operations. It will scan your documents and create classes for each operation.
    The class will have a `document` attribute that contains the query document, as well as contained "Arguments" class
    with the variables for the operation.

    This allows for the serialization of values in both directions.

    If you want to generate python functions instead, use the `funcs` plugin in ADDITION to this plugin.

    """

    config: OperationsPluginConfig = Field(default_factory=OperationsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:
        plugin_tree = []

        documents = parse_documents(
            client_schema, self.config.operations_glob or config.documents
        )

        definitions = documents.definitions
        operations = [
            node for node in definitions if isinstance(node, OperationDefinitionNode)
        ]

        for operation in operations:
            plugin_tree += generate_operation(
                operation, client_schema, config, self.config, registry
            )

        return plugin_tree



from pydantic import Field
from turms.processors.base import Processor, ProcessorConfig
from turms.config import GeneratorConfig


class IsortProcessorConfig(ProcessorConfig):
    type: str = "turms.processors.isort.IsortProcessor"


class IsortProcessor(Processor):
    """A processor that uses isort to format the generated python code
    imports in a specific order.

    """

    config: IsortProcessorConfig = Field(default_factory=IsortProcessorConfig)

    def run(self, gen_file: str, config: GeneratorConfig):
        import isort

        return isort.code(gen_file)



from pydantic import Field
from turms.processors.base import Processor, ProcessorConfig
from turms.config import GeneratorConfig


class DisclaimerProcessorConfig(ProcessorConfig):
    type: str = "turms.processors.black.DisclaimerProcessor"
    disclaimer: str = '''""" This file was code generated by turms. If you want to change the contents of this file, you should make sure to add the MergeProcessor to your config will keep your changes when you re-run turms)."""
    '''


class DisclaimerProcessor(Processor):
    """A processor that uses black to format the generated python code

    Black is a code formatter that is used to enforce a consistent style on the generated python code.
    It needs to be seperately installed via 'pip install black'.
    """

    config: DisclaimerProcessorConfig = Field(default_factory=DisclaimerProcessorConfig)

    def run(self, gen_file: str, config: GeneratorConfig):
        return self.config.disclaimer + "\n" + gen_file



from turms.parsers.base import Parser, ParserConfig
from typing import List
import ast
from pydantic_settings import SettingsConfigDict
from pydantic import Field, field_validator, ConfigDict


class PolyfillPluginConfig(ParserConfig):
    model_config = SettingsConfigDict(env_prefix="TURMS_PARSERS_POLYFILL_")
    type: str = "turms.parsers.polyfill.PolyfillPlugin"
    skip_underscore: bool = True
    python_version: str = "3.9"

    @field_validator("python_version", mode="before")
    def validate_python_version(cls, value):
        if isinstance(value, (int, float)):
            value = str(value)

        if value not in ["3.7", "3.8", "3.9", "3.10", "3.11", "3.12", "3.13"]:
            raise ValueError("Invalid python version")

        return value


def polyfill_python_seven(
    asts: List[ast.AST], config: PolyfillPluginConfig
) -> List[ast.AST]:

    new_nodes = []
    for node in asts:
        if isinstance(node, ast.ImportFrom):
            if node.module == "typing":
                # okay imports
                aliases = {alias.name for alias in node.names}
                need_extensions = aliases.intersection({"Literal"})
                if need_extensions:
                    new_nodes.append(
                        ast.ImportFrom(
                            module="typing_extensions",
                            names=[ast.alias(name=name) for name in need_extensions],
                            level=0,
                        )
                    )

                okay_imports = aliases - need_extensions
                new_nodes.append(
                    ast.ImportFrom(
                        names=[ast.alias(name=name) for name in okay_imports],
                        module="typing",
                        level=0,
                    )
                )
            else:
                new_nodes.append(node)

        else:
            new_nodes.append(node)

    return new_nodes


class PolyfillParser(Parser):
    """The polyfill parser is used to polyfill the generated python code with
    additional imports and code to make it compatible with older python versions.

    Right now it only supports polyfils for python 3.7 and higher"""

    config: PolyfillPluginConfig = Field(default_factory=PolyfillPluginConfig)

    def parse_ast(
        self,
        asts: List[ast.AST],
    ) -> List[ast.AST]:

        if self.config.python_version == "3.7":
            return polyfill_python_seven(asts, self.config)

        return asts  # pragma: no cover



from abc import abstractmethod
import ast
from typing import List

from pydantic import BaseModel, ConfigDict, Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.config import LogFunction


class PluginConfig(BaseSettings):
    model_config = SettingsConfigDict(extra="forbid")
    type: str


class Plugin(BaseModel):
    """
    Base class for all plugins

    Plugins are the workhorse of turms. They are used to generate python code, according
    to the GraphQL schema. You can use plugins to generate python code for your GraphQL
    schema. THe all received the graphql schema and the config of the plugin."""

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)
    config: PluginConfig
    log: LogFunction = Field(default=lambda *args, **kwargs: print(*args))

    @abstractmethod
    def generate_ast(
        self,
        config,
        client_schema: GraphQLSchema,
        registry,
    ) -> List[ast.AST]: ...  # pragma: no cover



from pydantic import Field
from turms.processors.base import Processor, ProcessorConfig
from turms.config import GeneratorConfig


class BlackProcessorConfig(ProcessorConfig):
    type: str = "turms.processors.black.BlackProcessor"


class BlackProcessor(Processor):
    """A processor that uses black to format the generated python code

    Black is a code formatter that is used to enforce a consistent style on the generated python code.
    It needs to be seperately installed via 'pip install black'.
    """

    config: BlackProcessorConfig = Field(default_factory=BlackProcessorConfig)

    def run(self, gen_file: str, config: GeneratorConfig):
        from black import format_str, FileMode

        return format_str(gen_file, mode=FileMode())



from pydantic_settings import SettingsConfigDict
from turms.plugins.base import Plugin, PluginConfig
import ast
from typing import List
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.plugins.base import Plugin
from pydantic import Field
from graphql.type.definition import (
    GraphQLEnumType,
)
from turms.referencer import create_reference_registry_from_documents
from turms.utils import parse_documents
from turms.registry import ClassRegistry


class EnumsPluginsError(Exception):
    pass


class EnumsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(env_prefix="TURMS_PLUGINS_ENUMS_")
    type: str = "turms.plugins.enums.EnumsPlugin"
    skip_underscore: bool = False
    skip_double_underscore: bool = True
    skip_unreferenced: bool = True
    prepend: str = ""
    append: str = ""


def generate_enums(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: EnumsPluginConfig,
    registry: ClassRegistry,
):
    tree = []

    enum_types = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLEnumType)
    }

    if plugin_config.skip_unreferenced and config.documents:
        ref_registry = create_reference_registry_from_documents(
            client_schema, parse_documents(client_schema, config.documents)
        )
    else:
        ref_registry = None

    for key, type in enum_types.items():
        if ref_registry and key not in ref_registry.enums:
            continue

        if plugin_config.skip_underscore and key.startswith("_"):
            continue

        if plugin_config.skip_double_underscore and key.startswith("__"):
            continue

        classname = registry.generate_enum(key)

        fields = (
            [ast.Expr(value=ast.Constant(value=type.description))]
            if type.description
            else []
        )

        for value_key, value in type.values.items():
            if isinstance(value.value, str):
                servalue = value.value
            else:
                servalue = value.value.value

            assign = ast.Assign(
                targets=[ast.Name(id=str(value_key), ctx=ast.Store())],
                value=ast.Constant(value=servalue),
            )

            potential_comment = (
                value.description
                if not value.deprecation_reason
                else f"DEPRECATED: {value.description}"
            )

            if potential_comment:
                fields += [
                    assign,
                    ast.Expr(value=ast.Constant(value=potential_comment)),
                ]

            else:
                fields += [assign]

        tree.append(
            ast.ClassDef(
                classname,
                bases=[
                    ast.Name(id="str", ctx=ast.Load()),
                    ast.Name(id="Enum", ctx=ast.Load()),
                ],
                decorator_list=[],
                keywords=[],
                body=fields,
            )
        )

    return tree


class EnumsPlugin(Plugin):
    """The Enum plugin generates python enums from the GraphQL schema.

    It does not need documents but generates every enum from the loaded schema.

    By providing a config, you can skip enums that start with an underscore and
    prepend and append strings to the generated enums."""

    config: EnumsPluginConfig = Field(default_factory=EnumsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:
        registry.register_import("enum.Enum")

        return generate_enums(client_schema, config, self.config, registry)



import ast
from typing import List, Optional

from pydantic_settings import SettingsConfigDict
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.recurse import type_field_node
from turms.plugins.base import Plugin, PluginConfig
from pydantic import Field
from graphql.language.ast import FragmentDefinitionNode
from turms.registry import ClassRegistry
from turms.utils import (
    generate_pydantic_config,
    generate_typename_field,
    get_additional_bases_for_type,
    get_interface_bases,
    parse_documents,
)
from graphql import (
    FieldNode,
    FragmentSpreadNode,
    GraphQLInterfaceType,
    GraphQLObjectType,
    InlineFragmentNode,
    language,
)
from turms.config import GraphQLTypes

from graphql.utilities.type_info import get_field_def
import logging


logger = logging.getLogger(__name__)


class FragmentsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(env_prefix="TURMS_PLUGINS_FRAGMENTS_")
    type: str = "turms.plugins.fragments.FragmentsPlugin"
    fragment_bases: List[str] = []
    fragments_glob: Optional[str] = None


def get_fragment_bases(
    config: GeneratorConfig,
    pluginConfig: FragmentsPluginConfig,
    registry: ClassRegistry,
):
    if pluginConfig.fragment_bases:
        for base in pluginConfig.fragment_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in pluginConfig.fragment_bases
        ]

    else:
        for base in config.object_bases:
            registry.register_import(base)

        return [
            ast.Name(id=base.split(".")[-1], ctx=ast.Load())
            for base in config.object_bases
        ]


def generate_fragment(
    f: FragmentDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FragmentsPluginConfig,
    registry: ClassRegistry,
):
    tree = []
    fields = []
    type = client_schema.get_type(f.type_condition.name.value)
    name = registry.generate_fragment(
        f.name.value, isinstance(type, GraphQLInterfaceType)
    )

    registry.register_fragment_document(
        f.name.value, language.print_ast(f)
    )  # TODO: Check if typename is being referenced? so that we can check between the elements of the interface

    if isinstance(type, GraphQLInterfaceType):
        mother_class_fields = []
        base_fragment_name = f"{name}"
        additional_bases = get_additional_bases_for_type(type.name, config, registry)

        if type.description:
            mother_class_fields.append(
                ast.Expr(value=ast.Constant(value=type.description))
            )

        for sub_node in f.selection_set.selections:

            if isinstance(sub_node, FieldNode):

                if sub_node.name.value == "__typename":
                    continue

                field_type = type.fields[sub_node.name.value]
                mother_class_fields += type_field_node(
                    sub_node,
                    base_fragment_name,
                    field_type,
                    client_schema,
                    config,
                    tree,
                    registry,
                )

        mother_class = ast.ClassDef(
            base_fragment_name,
            bases=get_interface_bases(config, registry)
            + additional_bases,  # Todo: fill with base
            decorator_list=[],
            keywords=[],
            body=mother_class_fields if mother_class_fields else [ast.Pass()],
        )

        tree.append(mother_class)

        union_class_names = []

        for sub_node in f.selection_set.selections:

            if isinstance(sub_node, FragmentSpreadNode):
                # Spread nodes are like inheritance?
                spreaded_name = f"{base_fragment_name}{sub_node.name.value}"

                cls = ast.ClassDef(
                    spreaded_name,
                    bases=[
                        ast.Name(
                            id=registry.inherit_fragment(sub_node.name.value),
                            ctx=ast.Load(),
                        ),
                        ast.Name(id=base_fragment_name, ctx=ast.Load()),
                    ],
                    decorator_list=[],
                    keywords=[],
                    body=[ast.Pass()],
                )

                tree.append(cls)
                union_class_names.append(spreaded_name)

            if isinstance(sub_node, InlineFragmentNode):
                inline_name = (
                    f"{base_fragment_name}{sub_node.type_condition.name.value}Fragment"
                )
                inline_fragment_fields = []

                inline_fragment_fields += [
                    generate_typename_field(
                        sub_node.type_condition.name.value, registry, config
                    )
                ]

                for sub_sub_node in sub_node.selection_set.selections:

                    if isinstance(sub_sub_node, FieldNode):
                        sub_sub_node_type = client_schema.get_type(
                            sub_node.type_condition.name.value
                        )

                        if sub_sub_node.name.value == "__typename":
                            continue

                        field_type = sub_sub_node_type.fields[sub_sub_node.name.value]
                        inline_fragment_fields += type_field_node(
                            sub_sub_node,
                            inline_name,
                            field_type,
                            client_schema,
                            config,
                            tree,
                            registry,
                        )

                additional_bases = get_additional_bases_for_type(
                    sub_node.type_condition.name.value, config, registry
                )
                cls = ast.ClassDef(
                    inline_name,
                    bases=[
                        ast.Name(id=base_fragment_name, ctx=ast.Load()),
                    ]
                    + additional_bases,
                    decorator_list=[],
                    keywords=[],
                    body=inline_fragment_fields
                    + generate_pydantic_config(
                        GraphQLTypes.FRAGMENT,
                        config,
                        registry,
                        sub_node.type_condition.name.value,
                    ),
                )

                tree.append(cls)
                union_class_names.append(inline_name)

        union_class_names.append(base_fragment_name)

        if len(union_class_names) > 1:
            registry.register_import("typing.Union")
            slice = ast.Tuple(
                elts=[
                    ast.Name(id=clsname, ctx=ast.Load())
                    for clsname in union_class_names
                ],
                ctx=ast.Load(),
            )
            tree.append(
                ast.Assign(
                    targets=[
                        ast.Name(
                            id=registry.style_fragment_class(f.name.value),
                            ctx=ast.Store(),
                        )
                    ],
                    value=ast.Subscript(
                        value=ast.Name("Union", ctx=ast.Load()),
                        slice=slice,
                        ctx=ast.Load(),
                    ),
                )
            )

        return tree

    elif isinstance(type, GraphQLObjectType):
        additional_bases = get_additional_bases_for_type(
            f.type_condition.name.value, config, registry
        )

        fields += [generate_typename_field(type.name, registry, config)]

        for field in f.selection_set.selections:

            if field.name.value == "__typename":
                continue

            if isinstance(field, FragmentDefinitionNode):  # pragma: no cover

                continue

            if isinstance(field, FragmentSpreadNode):
                additional_bases = [
                    ast.Name(
                        id=registry.inherit_fragment(field.name.value),
                        ctx=ast.Load(),
                    )
                ] + additional_bases  # needs to be prepended (MRO)
                continue

            field_definition = get_field_def(client_schema, type, field)
            assert field_definition, "Couldn't find field definition"

            fields += type_field_node(
                field,
                name,
                field_definition,
                client_schema,
                config,
                tree,
                registry,
            )

        tree.append(
            ast.ClassDef(
                name,
                bases=additional_bases
                + get_fragment_bases(config, plugin_config, registry),
                decorator_list=[],
                keywords=[],
                body=fields
                + generate_pydantic_config(GraphQLTypes.FRAGMENT, config, registry),
            )
        )
        return tree


class FragmentsPlugin(Plugin):
    """Plugin for generating fragments from
    documents

    The fragments plugin will generate classes for each fragment. It loads the documents,
    scans for fragments and generates the classes.

    If encountering a fragment on an interface it will generate a BASE class for that interface
    and then generate a class for each type referenced in the fragment. They will all inherit
    from the base class. The true type will be determined at runtime as all of the potential subtypes
    will be in the same union.

    """

    config: FragmentsPluginConfig = Field(default_factory=FragmentsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:

        plugin_tree = []

        documents = parse_documents(
            client_schema, self.config.fragments_glob or config.documents
        )

        definitions = documents.definitions

        fragments = [
            node for node in definitions if isinstance(node, FragmentDefinitionNode)
        ]

        for fragment in fragments:
            plugin_tree += generate_fragment(
                fragment, client_schema, config, self.config, registry
            )

        return plugin_tree



import ast
from keyword import iskeyword
from typing import Dict, List

from turms.config import GeneratorConfig, LogFunction
from turms.errors import (
    NoEnumFound,
    NoInputTypeFound,
    NoScalarFound,
    RegistryError,
)
from turms.stylers.base import Styler

SCALAR_DEFAULTS = {
    "ID": "str",
    "String": "str",
    "Int": "int",
    "Boolean": "bool",
    "GenericScalar": "typing.Dict",
    "DateTime": "datetime.datetime",
    "Float": "float",
}


built_in_map = {
    "__TypeKind": ast.ClassDef(
        "__TypeKind",
        bases=[
            ast.Name(id="str", ctx=ast.Load()),
            ast.Name(id="Enum", ctx=ast.Load()),
        ],
        decorator_list=[],
        keywords=[],
        body=[
            ast.Assign(
                targets=[ast.Name(id=str(value), ctx=ast.Store())],
                value=ast.Constant(value=value),
            )
            for value in ["OBJECT", "INTERFACE", "UNION", "ENUM", "INPUT_OBJECT"]
        ],
    ),
    "__DirectiveLocation": ast.ClassDef(
        "__DirectiveLocation",
        bases=[
            ast.Name(id="str", ctx=ast.Load()),
            ast.Name(id="Enum", ctx=ast.Load()),
        ],
        decorator_list=[],
        keywords=[],
        body=[
            ast.Assign(
                targets=[ast.Name(id=str(value), ctx=ast.Store())],
                value=ast.Constant(value=value),
            )
            for value in [
                "SCALAR",
                "OBJECT",
                "FIELD_DEFINITION",
                "ARGUMENT_DEFINITION",
                "INTERFACE",
                "UNION",
                "ENUM",
                "ENUM_VALUE",
                "INPUT_OBJECT",
                "INPUT_FIELD_DEFINITION",
                "SCHEMA",
            ]
        ],
    ),
}  # builtin map provides the default types for any schema if they are referenced


class ClassRegistry(object):
    """Class Registry is responsible for keeping track of all the classes that are generated
    as well as their names. It also keeps track of all the imports that are required for the
    generated code to work, as well as all the forward references that are required
    for the generated code to work (i.e. when a class references another class that has not
    yet been defined). Class Registry provides a facade to the rest of the code to abstract
    the logic behind the stylers."""

    def __init__(
        self, config: GeneratorConfig, stylers: List[Styler], log: LogFunction
    ):
        self.stylers = stylers
        self._imports = set()
        self._builtins = set()
        self.config = config

        self.scalar_map = {**SCALAR_DEFAULTS, **config.scalar_definitions}

        self.fragment_document_map = {}

        self.enum_class_map = {}
        self.inputtype_class_map = {}
        self.object_class_map = {}
        self.interface_reference_map = {}
        self.interface_baseclass_map = {}

        self.operation_class_map = {}
        self.fragment_class_map = {}
        self.query_class_map = {}
        self.subscription_class_map = {}
        self.mutation_class_map = {}

        self.registered_interfaces_fragments = {}
        self.registered_union_fragments = {}
        self.forward_references = set()
        self.fragment_type_map = {}

        self.interfacefragments_impl_map = {}
        self.unionfragment_members_map = {}
        self.log = log

    def style_inputtype_class(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_input_name(typename)
        if iskeyword(typename):
            typename += "_"
        return typename

    def generate_inputtype(self, typename: str):
        assert (
            typename not in self.inputtype_class_map
        ), "Type was already registered, cannot register annew"
        classname = self.style_inputtype_class(typename)
        self.inputtype_class_map[typename] = classname
        return classname

    def get_inputtype_class(self, typename) -> str:
        return self.inputtype_class_map[typename]

    def reference_inputtype(
        self, typename: str, parent: str, allow_forward=True
    ) -> ast.AST:
        classname = self.style_inputtype_class(typename)
        if typename not in self.inputtype_class_map or parent == classname:
            if not allow_forward:
                raise NoInputTypeFound(
                    f"Input type {typename} is not yet defined but referenced by {parent}. And we dont allow forward references"
                )
            self.forward_references.add(parent)
            return ast.Constant(value=classname, ctx=ast.Load())
        return ast.Name(id=self.inputtype_class_map[typename], ctx=ast.Load())

    def style_enum_class(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_enum_name(typename)
        if iskeyword(typename):
            typename += "_"
        return typename

    def generate_enum(self, typename: str):
        assert (
            typename not in self.enum_class_map
        ), "Type was already registered, cannot register annew"
        classname = self.style_enum_class(typename)
        self.enum_class_map[typename] = classname
        return classname

    def get_enum_class(self, typename) -> str:
        return self.enum_class_map[typename]

    def reference_enum(self, typename: str, parent: str, allow_forward=True) -> ast.AST:
        if typename in built_in_map:
            # Builtin enums
            self._builtins.add(typename)
            self.forward_references.add(parent)
            return ast.Constant(value=typename, ctx=ast.Load())

        classname = self.style_enum_class(typename)
        if typename not in self.enum_class_map or parent == classname:
            if not allow_forward:
                raise NoEnumFound(
                    f"Input type {typename} is not yet defined but referenced by {parent}. And we dont allow forward references"
                )
            self.forward_references.add(parent)
            return ast.Constant(value=classname, ctx=ast.Load())
        return ast.Name(id=classname, ctx=ast.Load())

    def style_objecttype_class(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_object_name(typename)
        if iskeyword(typename):
            typename += "_"
        return typename

    def generate_objecttype(self, typename: str):
        assert (
            typename not in self.object_class_map
        ), "Type was already registered, cannot register annew"
        classname = self.style_objecttype_class(typename)
        self.object_class_map[typename] = classname
        return classname

    def reference_object(
        self, typename: str, parent: str, allow_forward=True
    ) -> ast.AST:
        return self._reference_generic(
            typename,
            parent,
            self.style_objecttype_class,
            self.object_class_map,
            "Object",
            allow_forward,
        )

    def _reference_generic(
        self,
        typename: str,
        parent: str,
        styler,
        class_map: Dict,
        error_class: str,
        allow_forward=True,
    ) -> ast.AST:
        classname = styler(typename)
        if typename not in class_map or parent == classname:
            if not allow_forward:
                raise RegistryError(
                    f"""{error_class} type {typename} is not yet defined but referenced by {parent}. And we dont allow forward references"""
                )
            self.forward_references.add(parent)
            return ast.Constant(value=classname, ctx=ast.Load())
        return ast.Name(id=classname, ctx=ast.Load())

    def style_interface_class(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_object_name(typename)
        if iskeyword(typename):
            typename += "_"
        return typename

    def generate_interface(self, typename: str, with_base: bool = True):
        assert (
            typename not in self.interface_baseclass_map
        ), "Type was already registered, cannot register annew"
        classname = self.style_interface_class(typename)
        if with_base:
            self.interface_baseclass_map[typename] = classname + "Base"
        else:
            self.interface_baseclass_map[typename] = classname
        self.interface_reference_map[typename] = classname
        return self.interface_baseclass_map[typename]

    def inherit_interface(self, typename: str, allow_forward=True) -> ast.AST:
        if typename not in self.interface_baseclass_map:
            raise RegistryError(
                f"Interface type {typename} is not yet defined but referenced. Please define it first."
            )
        return self.interface_baseclass_map[typename]

    def reference_interface(
        self, typename: str, parent: str, allow_forward=True
    ) -> ast.AST:
        # Interface need always be udpated later
        classname = self.style_interface_class(typename)
        self.forward_references.add(parent)
        return ast.Constant(value=classname, ctx=ast.Load())

    def style_fragment_class(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_fragment_name(typename)
        if iskeyword(typename):
            typename += "_"
        return typename

    def generate_fragment(self, fragmentname: str, is_interface=False):
        assert (
            fragmentname not in self.fragment_class_map
        ), f"Fragment {fragmentname} was already registered, cannot register annew"
        classname = self.style_fragment_class(fragmentname)
        real_classname = classname if not is_interface else classname
        self.fragment_class_map[fragmentname] = real_classname
        return real_classname

    def register_fragment_type(self, fragmentname: str, typename: str):
        self.fragment_type_map[fragmentname] = typename


    def register_interface_fragment_implementations(self, fragmentname: str, implementationMap: Dict[str, str]):
        self.interfacefragments_impl_map[fragmentname] = implementationMap


    def get_interface_fragment_implementations(self, fragmentname: str):
        return self.interfacefragments_impl_map[fragmentname]


    def register_union_fragment_members(self, fragmentname: str, membersMap: Dict[str, str]):
        self.unionfragment_members_map[fragmentname] = membersMap


    def get_union_fragment_members(self, fragmentname: str):
        return self.unionfragment_members_map[fragmentname]


    def get_fragment_type(self, fragmentname: str):
        return self.fragment_type_map[fragmentname]



    def reference_fragment(
        self, typename: str, parent: str, allow_forward=True
    ) -> ast.AST:
        return self._reference_generic(
            typename,
            parent,
            self.style_fragment_class,
            self.fragment_class_map,
            "Fragment",
            allow_forward,
        )
    
    def is_interface_fragment(self, typename: str):
        return typename in self.registered_interfaces_fragments
    

    def reference_interface_fragment(self, typename: str, parent: str, allow_forward=True) -> ast.AST:
        return self.registered_interfaces_fragments[typename]
    
    def register_interface_fragment(self, typename: str, ast: ast.AST):
        self.registered_interfaces_fragments[typename] = ast

    def inherit_fragment(self, typename: str, allow_forward=True) -> ast.AST:
        if typename not in self.fragment_class_map:
            raise RegistryError(
                f"Fragment {typename} is not yet defined but referenced by. Please change the order in your fragments."
            )
        return self.fragment_class_map[typename]

    def generate_node_name(self, node_name: str):
        for styler in self.stylers:
            node_name = styler.style_node_name(node_name)

        if iskeyword(node_name):
            return node_name + "_"

        return node_name

    def generate_parameter_name(self, node_name: str):
        for styler in self.stylers:
            node_name = styler.style_parameter_name(node_name)

        if iskeyword(node_name):
            return node_name + "_"

        return node_name

    def style_query_class(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_query_name(typename)
        if iskeyword(typename):
            typename += "_"
        return typename

    def generate_query(self, typename: str):
        assert (
            typename not in self.query_class_map
        ), f"Type {typename} was already registered, cannot register annew"
        classname = self.style_query_class(typename)
        self.query_class_map[typename] = classname
        return classname

    def reference_query(
        self, typename: str, parent: str, allow_forward=True
    ) -> ast.AST:
        return self._reference_generic(
            typename,
            parent,
            self.style_query_class,
            self.query_class_map,
            "Query",
            allow_forward,
        )

    def style_mutation_class(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_mutation_name(typename)
        if iskeyword(typename):
            typename += "_"
        return typename

    def generate_mutation(self, typename: str):
        assert (
            typename not in self.mutation_class_map
        ), f"Type {typename} was already registered, cannot register annew"
        classname = self.style_mutation_class(typename)
        self.mutation_class_map[typename] = classname
        return classname

    def reference_mutation(
        self, typename: str, parent: str, allow_forward=True
    ) -> ast.AST:
        return self._reference_generic(
            typename,
            parent,
            self.style_mutation_class,
            self.mutation_class_map,
            "Mutation",
            allow_forward,
        )

    def style_subscription_class(self, typename: str):
        for styler in self.stylers:
            typename = styler.style_subscription_name(typename)
        if iskeyword(typename):
            typename += "_"
        return typename

    def generate_subscription(self, typename: str):
        assert (
            typename not in self.subscription_class_map
        ), f"Type {typename} was already registered, cannot register annew"
        classname = self.style_subscription_class(typename)
        self.subscription_class_map[typename] = classname
        return classname

    def reference_subscription(
        self, typename: str, parent: str, allow_forward=True
    ) -> ast.AST:
        return self._reference_generic(
            typename,
            parent,
            self.style_subscription_class,
            self.subscription_class_map,
            "Subscription",
            allow_forward,
        )

    def register_import(self, name):
        if name in ("bool", "str", "int", "float", "dict", "list", "tuple"):
            return

        self._imports.add(name)

    def generate_imports(self):
        imports = []

        lone_top_imports = set()
        top_level = {}
        for name in self._imports:
            if "." not in name:
                lone_top_imports.add(name)
            else:
                top_level.setdefault(".".join(name.split(".")[:-1]), set()).add(
                    name.split(".")[-1]
                )

        for lone_top_import in lone_top_imports:
            imports.append(ast.Import(names=[ast.alias(name=lone_top_import)]))

        for top_level_name, sub_level_names in top_level.items():
            imports.append(
                ast.ImportFrom(
                    module=top_level_name,
                    names=[ast.alias(name=name) for name in sub_level_names],
                    level=0,
                )
            )

        return imports

    def generate_builtins(self):
        builtins = []

        for built_in in self._builtins:
            builtins.append(built_in_map[built_in])

        return builtins

    def generate_forward_refs(self):
        tree = []

        for reference in sorted(self.forward_references):
            tree.append(
                ast.Expr(
                    value=ast.Call(
                        func=ast.Attribute(
                            value=ast.Name(
                                id=reference,
                                ctx=ast.Load(),
                            ),
                            attr=(
                                "model_rebuild"
                                if self.config.pydantic_version == "v2"
                                else "update_forward_refs"
                            ),
                            ctx=ast.Load(),
                        ),
                        keywords=[],
                        args=[],
                    )
                )
            )

        return tree

    def register_fragment_document(self, typename: str, cls: str):
        assert cls not in self.fragment_document_map, f"{cls} already registered"
        self.fragment_document_map[typename] = cls

    def get_fragment_document(self, typename: str):
        return self.fragment_document_map[typename]

    def register_scalar(self, scalar_type: str, python_type: str):
        self.scalar_map[scalar_type] = python_type

    def reference_scalar(self, scalar_type: str):
        if scalar_type in self.scalar_map:
            python_type = self.scalar_map[scalar_type]
            if "." in python_type:
                # We make the assumption that the scalar type is a fully qualified class name
                self.register_import(python_type)

            return ast.Name(
                id=python_type.split(".")[-1], ctx=ast.Load()
            )  # That results only in the class name (also in the case of a builtin)

        raise NoScalarFound(
            f"No python equivalent found for {scalar_type}. Please define in scalar_definitions"
        )

    def warn(self, message):
        self.log(message, level="WARN")



from graphql import (
    BooleanValueNode,
    ConstListValueNode,
    ConstObjectValueNode,
    ConstValueNode,
    EnumValueNode,
    FloatValueNode,
    GraphQLField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLType,
    GraphQLUnionType,
    IntValueNode,
    ListValueNode,
    NullValueNode,
    ObjectValueNode,
    StringValueNode,
    Undefined,
    GraphQLArgument,
    ObjectTypeDefinitionNode,
)
from pydantic_settings import SettingsConfigDict

from turms.plugins.base import Plugin, PluginConfig
import ast
from typing import Dict, List, Protocol, runtime_checkable
from turms.config import GeneratorConfig, ImportableFunctionMixin
from graphql.utilities.build_client_schema import GraphQLSchema
from pydantic import Field
from graphql.type.definition import (
    GraphQLEnumType,
)
from turms.registry import ClassRegistry
from turms.utils import (
    generate_pydantic_config,
    get_additional_bases_for_type,
    interface_is_extended_by_other_interfaces,
)
from turms.config import GraphQLTypes


@runtime_checkable
class StrawberryGenerateFunc(ImportableFunctionMixin, Protocol):
    def __call__(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        plugin_config: "StrawberryPluginConfig",
        registry: ClassRegistry,
    ) -> List[ast.AST]: ...  # pragma: no cover


def build_directive_type_annotation(value: GraphQLType, registry: ClassRegistry, is_optional=True):

    if isinstance(value, GraphQLScalarType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(value.name),
                ctx=ast.Load(),
            )
        
        return registry.reference_scalar(value.name)
    if isinstance(value, GraphQLObjectType):
        raise NotImplementedError("Object types cannot be used as arguments")
    if isinstance(value, GraphQLInterfaceType):
        raise NotImplementedError("Interface types cannot be used as arguments")
    if isinstance(value, GraphQLUnionType):
        raise NotImplementedError("Union types cannot be used as arguments")
    if isinstance(value, GraphQLEnumType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_enum(value.name),
                ctx=ast.Load(),
            )

        return registry.reference_enum(value.name)
    if isinstance(value, GraphQLNonNull):
        return build_directive_type_annotation(value.of_type, registry, is_optional=False)
    if isinstance(value, GraphQLList):
        registry.register_import("typing.List")

        if is_optional:
            registry.register_import("typing.Optional")

            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()),
                    slice=build_directive_type_annotation(value.of_type, registry, is_optional=True),
                    ctx=ast.Load(),
                ),
                ctx=ast.Load(),
            )

        return ast.Subscript(
            value=ast.Name("List", ctx=ast.Load()),
            slice=build_directive_type_annotation(value.of_type, registry, is_optional=True),
            ctx=ast.Load(),
        )
    if isinstance(value, GraphQLInputObjectType):
        raise NotImplementedError("Input types cannot be used as arguments")
    
    raise NotImplementedError(f"Unknown type {repr(value)}")



def convert_valuenode_to_ast(value: ConstValueNode):
    if isinstance(value, NullValueNode):
        return ast.Constant(value=None)
    if isinstance(value, StringValueNode):
        return ast.Constant(value=value.value)
    if isinstance(value, IntValueNode):
        return ast.Constant(value=value.value)
    if isinstance(value, FloatValueNode):
        return ast.Constant(value=value.value)
    if isinstance(value, BooleanValueNode):
        return ast.Constant(value=value.value)

    if isinstance(value, EnumValueNode):
        return ast.Constant(value=value)
    if isinstance(value, ListValueNode):
        return ast.List(elts=[convert_valuenode_to_ast(x) for x in value.values], ctx=ast.Load())
    if isinstance(value, ObjectValueNode):

        keys = []
        values = []

        for field in value.fields:
            keys.append(field.name.value)
            values.append(convert_valuenode_to_ast(field.value))

        return ast.Dict(
            keys=keys,
            values=values,
        )
    
    raise NotImplementedError(f"Unknown default value {repr(value)}")

    

def convert_default_value_to_ast(value):
    if value is Undefined:
        return None
    if value is None:
        return ast.Constant(value=None)
    if isinstance(value, str):
        return ast.Constant(value=value)
    if isinstance(value, int):
        return ast.Constant(value=value)
    if isinstance(value, float):
        return ast.Constant(value=value)
    if isinstance(value, bool):
        return ast.Constant(value=value)
    if isinstance(value, list):
        return ast.List(elts=[convert_default_value_to_ast(x) for x in value], ctx=ast.Load())
    if isinstance(value, dict):
        keys = []
        values = []

        for key, value in value.items():
            keys.append(key)
            values.append(convert_default_value_to_ast(value))

        return ast.Dict(
            keys=keys,
            values=values,
        )
    raise NotImplementedError(f"Unknown default value {repr(value)}")







def default_generate_directives(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: "StrawberryPluginConfig",
    registry: ClassRegistry,
):
    tree = []

    directives = client_schema.directives

    generatable_directives = [
        directive
        for directive in directives
        if directive.name not in plugin_config.builtin_directives
    ]

    if not generatable_directives:
        return []

    registry.register_import("strawberry.schema_directive.Location")

    for directive in generatable_directives:
        if plugin_config.skip_underscore and directive.name.startswith("_"):
            continue

        if plugin_config.skip_double_underscore and directive.name.startswith("__"):
            continue

        keywords = []
        fields = []

        keywords.append(
            ast.keyword(
                arg="locations",
                value=ast.List(
                    elts=[
                        ast.Attribute(
                            value=ast.Name(id="Location", ctx=ast.Load()),
                            attr=location.name,
                            ctx=ast.Load(),
                        )
                        for location in directive.locations
                    ],
                    ctx=ast.Load(),
                ),
            )
        )

        for value_key, value in directive.args.items():
            value: GraphQLArgument

            type = value.type

            if value.default_value is not None:
                default = convert_default_value_to_ast(value.default_value)
            else:
                default = None

            needs_factory = False
            if isinstance(default, ast.List):
                needs_factory = True
            if isinstance(default, ast.Dict):
                needs_factory = True


            field_value = None

            if default:
                if needs_factory:
                    field_value = ast.Call(
                        func=ast.Name(id="strawberry.field", ctx=ast.Load()),
                        keywords=[
                            ast.keyword(
                                arg="default_factory",
                                value=ast.Lambda(
                                    args=[], body=default
                                ),
                            ),
                        ],
                        args=[],
                    )
                else:
                    field_value = ast.Call(
                        func=ast.Name(id="strawberry.field", ctx=ast.Load()),
                        keywords=[
                            ast.keyword(
                                arg="default",
                                value=default,
                            ),
                        ],
                        args=[],
                    )


            assign = ast.AnnAssign(
                target=ast.Name(
                    id=registry.generate_node_name(value_key), ctx=ast.Store()
                ),
                annotation=build_directive_type_annotation(type, registry),
                value=field_value,
                simple=1,
            )

            fields += [assign]

        decorator = ast.Call(
            func=ast.Name(
                id="strawberry.schema_directive",
                ctx=ast.Load(),
            ),
            keywords=keywords,
            args=[],
        )

        tree.append(
            ast.ClassDef(
                name=directive.name,
                bases=[],
                decorator_list=[decorator],
                keywords=[],
                body=(fields or [ast.Pass()])
                + generate_pydantic_config(
                    GraphQLTypes.DIRECTIVE, config, registry, directive.name
                ),
            )
        )

    return tree


def default_generate_enums(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: "StrawberryPluginConfig",
    registry: ClassRegistry,
):
    tree = []

    enum_types = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLEnumType)
    }

    for key, type in enum_types.items():
        directive_keywords = generate_directive_keywords(type.ast_node, plugin_config)
        if directive_keywords:
            decorator = ast.Call(
                func=ast.Name(id="strawberry.enum", ctx=ast.Load()),
                keywords=directive_keywords,
                args=[],
            )
        else:
            decorator = ast.Name(id="strawberry.enum", ctx=ast.Load())

        if plugin_config.skip_underscore and key.startswith("_"):
            continue

        if plugin_config.skip_double_underscore and key.startswith("__"):
            continue

        classname = registry.generate_enum(key)

        fields = (
            [ast.Expr(value=ast.Constant(value=type.description))]
            if type.description
            else []
        )

        for value_key, value in type.values.items():
            if isinstance(value.value, str):
                servalue = value.value
            else:
                servalue = value.value.value

            assign = ast.Assign(
                targets=[ast.Name(id=str(value_key), ctx=ast.Store())],
                value=ast.Constant(value=servalue),
            )

            potential_comment = (
                value.description
                if not value.deprecation_reason
                else f"DEPRECATED: {value.description}"
            )

            if potential_comment:
                fields += [
                    assign,
                    ast.Expr(value=ast.Constant(value=potential_comment)),
                ]

            else:
                fields += [assign]

        registry.register_import("enum.Enum")

        tree.append(
            ast.ClassDef(
                classname,
                bases=[
                    ast.Name(id="Enum", ctx=ast.Load()),
                ],
                decorator_list=[decorator],
                keywords=[],
                body=fields or [ast.Pass()],
            )
        )

    return tree


class StrawberryPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(env_prefix="TURMS_PLUGINS_STRAWBERRY_")
    type: str = "turms.plugins.strawberry.Strawberry"
    generate_directives: bool = True
    generate_scalars: bool = True
    builtin_directives: List[str] = ["include", "skip", "deprecated", "specifiedBy"]
    builtin_scalars: List[str] = ["String", "Boolean", "DateTime", "Int", "Float", "ID"]
    generate_enums: bool = True
    generate_types: bool = True
    generate_inputs: bool = True
    types_bases: List[str] = []
    inputtype_bases: List[str] = []
    skip_underscore: bool = False
    skip_double_underscore: bool = True

    # Functional configuration:
    generate_directives_func: StrawberryGenerateFunc = default_generate_directives
    generate_enums_func: StrawberryGenerateFunc = default_generate_enums


def generate_object_field_annotation(
    graphql_type: GraphQLType,
    parent: str,
    config: GeneratorConfig,
    plugin_config: StrawberryPluginConfig,
    registry: ClassRegistry,
    is_optional=True,
):
    if isinstance(graphql_type, GraphQLScalarType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(graphql_type.name),
                ctx=ast.Load(),
            )

        return registry.reference_scalar(graphql_type.name)

    if isinstance(graphql_type, GraphQLInterfaceType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_interface(graphql_type.name, parent),
                ctx=ast.Load(),
            )

        return registry.reference_interface(graphql_type.name, parent)

    if isinstance(graphql_type, GraphQLObjectType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_object(graphql_type.name, parent),
                ctx=ast.Load(),
            )
        return registry.reference_object(graphql_type.name, parent)

    if isinstance(graphql_type, GraphQLUnionType):
        if is_optional:
            registry.register_import("typing.Optional")
            registry.register_import("typing.Union")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name("Union", ctx=ast.Load()),
                    slice=ast.Tuple(
                        elts=[
                            generate_object_field_annotation(
                                union_type,
                                parent,
                                config,
                                plugin_config,
                                registry,
                                is_optional=False,
                            )
                            for union_type in graphql_type.types
                        ],
                        ctx=ast.Load(),
                    ),
                ),
                ctx=ast.Load(),
            )
        registry.register_import("typing.Union")

        return ast.Subscript(
            value=ast.Name("Union", ctx=ast.Load()),
            slice=ast.Tuple(
                elts=[
                    generate_object_field_annotation(
                        union_type,
                        parent,
                        config,
                        plugin_config,
                        registry,
                        is_optional=False,
                    )
                    for union_type in graphql_type.types
                ],
                ctx=ast.Load(),
            ),
            ctx=ast.Load(),
        )

    if isinstance(graphql_type, GraphQLEnumType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_enum(
                    graphql_type.name,
                    parent,
                    allow_forward=not config.force_plugin_order,
                ),
                ctx=ast.Load(),
            )
        return registry.reference_enum(
            graphql_type.name, parent, allow_forward=not config.force_plugin_order
        )

    if isinstance(graphql_type, GraphQLNonNull):
        return generate_object_field_annotation(
            graphql_type.of_type,
            parent,
            config,
            plugin_config,
            registry,
            is_optional=False,
        )

    if isinstance(graphql_type, GraphQLList):
        registry.register_import("typing.List")

        def list_builder(x):
            return ast.Subscript(
                value=ast.Name("List", ctx=ast.Load()), slice=x, ctx=ast.Load()
            )

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=list_builder(
                    generate_object_field_annotation(
                        graphql_type.of_type,
                        parent,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    )
                ),
                ctx=ast.Load(),
            )

        registry.register_import("typing.List")
        return list_builder(
            generate_object_field_annotation(
                graphql_type.of_type,
                parent,
                config,
                plugin_config,
                registry,
                is_optional=True,
            )
        )

    raise NotImplementedError(
        f"Unknown object type {repr(graphql_type)}"
    )  # pragma: no cover


def recurse_argument_annotation(
    graphql_type: GraphQLType,
    parent: str,
    config: GeneratorConfig,
    plugin_config: StrawberryPluginConfig,
    registry: ClassRegistry,
    is_optional=True,
):
    if isinstance(graphql_type, GraphQLScalarType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(graphql_type.name),
                ctx=ast.Load(),
            )

        return registry.reference_scalar(graphql_type.name)

    if isinstance(graphql_type, GraphQLInputObjectType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_inputtype(graphql_type.name, parent),
                ctx=ast.Load(),
            )
        return registry.reference_inputtype(graphql_type.name, parent)

    if isinstance(graphql_type, GraphQLUnionType):
        if is_optional:
            registry.register_import("typing.Optional")
            registry.register_import("typing.Union")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name("Union", ctx=ast.Load()),
                    slice=ast.Tuple(
                        elts=[
                            recurse_argument_annotation(
                                union_type,
                                parent,
                                config,
                                plugin_config,
                                registry,
                                is_optional=False,
                            )
                            for union_type in graphql_type.types
                        ],
                        ctx=ast.Load(),
                    ),
                ),
                ctx=ast.Load(),
            )
        registry.register_import("typing.Union")

        return ast.Subscript(
            value=ast.Name("Union", ctx=ast.Load()),
            slice=ast.Tuple(
                elts=[
                    generate_object_field_annotation(
                        union_type,
                        parent,
                        config,
                        plugin_config,
                        registry,
                        is_optional=False,
                    )
                    for union_type in graphql_type.types
                ],
                ctx=ast.Load(),
            ),
            ctx=ast.Load(),
        )

    if isinstance(graphql_type, GraphQLEnumType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_enum(
                    graphql_type.name,
                    parent,
                    allow_forward=not config.force_plugin_order,
                ),
                ctx=ast.Load(),
            )
        return registry.reference_enum(
            graphql_type.name, parent, allow_forward=not config.force_plugin_order
        )

    if isinstance(graphql_type, GraphQLNonNull):
        return recurse_argument_annotation(
            graphql_type.of_type,
            parent,
            config,
            plugin_config,
            registry,
            is_optional=False,
        )

    if isinstance(graphql_type, GraphQLList):
        registry.register_import("typing.List")

        def list_builder(x):
            return ast.Subscript(
                value=ast.Name("List", ctx=ast.Load()), slice=x, ctx=ast.Load()
            )

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=list_builder(
                    recurse_argument_annotation(
                        graphql_type.of_type,
                        parent,
                        config,
                        plugin_config,
                        registry,
                        is_optional=True,
                    )
                ),
                ctx=ast.Load(),
            )

        registry.register_import("typing.List")
        return list_builder(
            recurse_argument_annotation(
                graphql_type.of_type,
                parent,
                config,
                plugin_config,
                registry,
                is_optional=True,
            )
        )

    raise NotImplementedError(f"Unknown input type {repr(graphql_type)}")


def generate_directive_keywords(
    ast_node: ObjectTypeDefinitionNode, plugin_config: StrawberryPluginConfig
):
    if not ast_node:
        return []

    directives = [
        directive
        for directive in ast_node.directives
        if directive.name.value not in plugin_config.builtin_directives
    ]

    if directives:
        calls = [
            ast.Call(
                func=ast.Name(
                    id=directive.name.value,
                    ctx=ast.Load(),
                ),
                keywords=[
                    ast.keyword(arg=arg.name.value, value=convert_valuenode_to_ast(arg.value))
                    for arg in directive.arguments
                ],
                args=[],
            )
            for directive in directives
        ]

        return [
            ast.keyword(
                arg="directives",
                value=ast.List(elts=calls, ctx=ast.Load()),
            )
        ]

    return []


def generate_inputs(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: StrawberryPluginConfig,
    registry: ClassRegistry,
):
    tree = []

    inputobjects_type = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLInputObjectType)
    }

    for key, type in inputobjects_type.items():
        if plugin_config.skip_underscore and key.startswith("_"):  # pragma: no cover
            continue

        directive_keywords = generate_directive_keywords(type.ast_node, plugin_config)
        if directive_keywords:
            decorator = ast.Call(
                func=ast.Name(id="strawberry.input", ctx=ast.Load()),
                keywords=directive_keywords,
                args=[],
            )
        else:
            decorator = ast.Name(id="strawberry.input", ctx=ast.Load())

        additional_bases = get_additional_bases_for_type(type.name, config, registry)
        name = registry.generate_inputtype(key)
        fields = (
            [ast.Expr(value=ast.Constant(value=type.description))]
            if type.description
            else []
        )

        for value_key, value in type.fields.items():
            value: GraphQLField

            keywords = []
            if value.description:
                keywords.append(
                    ast.keyword(
                        arg="description",
                        value=ast.Constant(value=value.description),
                    )
                )

            if value.deprecation_reason:
                keywords.append(
                    ast.keyword(
                        arg="deprecation_reason",
                        value=ast.Constant(value=value.deprecation_reason),
                    )
                )

            if keywords:
                assign_value = ast.Call(
                    func=ast.Name(
                        id="strawberry.field",
                        ctx=ast.Load(),
                    ),
                    keywords=keywords,
                    args=[],
                )
            else:
                assign_value = None

            assign = ast.AnnAssign(
                target=ast.Name(
                    registry.generate_node_name(value_key), ctx=ast.Store()
                ),
                annotation=recurse_argument_annotation(
                    value.type,
                    name,
                    config,
                    plugin_config,
                    registry,
                    is_optional=True,
                ),
                value=assign_value,
                simple=1,
            )

            fields += [assign]

        tree.append(
            ast.ClassDef(
                name,
                bases=additional_bases
                + [
                    ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                    for base in plugin_config.inputtype_bases
                ],
                decorator_list=[decorator],
                keywords=[],
                body=fields
                + generate_pydantic_config(
                    GraphQLTypes.INPUT, config, registry, typename=key
                ),
            )
        )

    return tree


def generate_types(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: StrawberryPluginConfig,
    registry: ClassRegistry,
):
    tree = []

    objects = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLObjectType)
        and not isinstance(value, GraphQLInputObjectType)
        or isinstance(value, GraphQLInterfaceType)
    }

    sorted_objects = {
        k: v
        for k, v in sorted(
            objects.items(),
            key=lambda item: isinstance(item[1], GraphQLInterfaceType),
            reverse=True,
        )
    }
    registry.register_import("strawberry")

    interface_map: Dict[str, List[str]] = (
        {}
    )  # A list of interfaces with the union classes attached
    interface_base_map: Dict[str, str] = (
        {}
    )  # A list of interfaces with its respective base

    for key, object_type in sorted_objects.items():
        additional_bases = get_additional_bases_for_type(
            object_type.name, config, registry
        )

        if plugin_config.skip_underscore and key.startswith("_"):
            continue

        if plugin_config.skip_double_underscore and key.startswith("__"):
            continue

        directive_keywords = generate_directive_keywords(
            object_type.ast_node, plugin_config
        )

        if isinstance(object_type, GraphQLObjectType):
            classname = registry.generate_objecttype(key)
            decorator_name = "strawberry.type"
        if isinstance(object_type, GraphQLInterfaceType):
            classname = registry.generate_interface(key, with_base=False)
            decorator_name = "strawberry.interface"
            interface_base_map[key] = classname

        if directive_keywords:
            decorator = ast.Call(
                func=ast.Name(id=decorator_name, ctx=ast.Load()),
                keywords=directive_keywords,
                args=[],
            )
        else:
            decorator = ast.Name(id=decorator_name, ctx=ast.Load())

        for interface in object_type.interfaces:
            # Populate the Union_classed
            interface_map.setdefault(interface.name, []).append(classname)

            other_interfaces = set(object_type.interfaces) - {interface}
            if not interface_is_extended_by_other_interfaces(
                interface, other_interfaces
            ):
                additional_bases.append(
                    ast.Name(
                        id=registry.inherit_interface(interface.name),
                        ctx=ast.Load(),
                    )
                )

        fields = (
            [ast.Expr(value=ast.Constant(value=object_type.description))]
            if object_type.description
            else []
        )

        for value_key, value in object_type.fields.items():
            value: GraphQLField

            keywords = []

            if value.description:
                keywords.append(
                    ast.keyword(
                        arg="description",
                        value=ast.Constant(value=value.description),
                    )
                )

            if value.deprecation_reason:
                keywords.append(
                    ast.keyword(
                        arg="deprecation_reason",
                        value=ast.Constant(value=value.deprecation_reason),
                    )
                )

            additional_args = []
            kwdefaults = []
            if value.args:
                sorted_args = {
                    key: value
                    for key, value in sorted(
                        value.args.items(),
                        key=lambda item: item[1].default_value == Undefined,
                        reverse=True,
                    )
                }

                for argkey, arg in sorted_args.items():
                    additional_args.append(
                        ast.arg(
                            arg=registry.generate_parameter_name(argkey),
                            annotation=recurse_argument_annotation(
                                arg.type,
                                classname,
                                config,
                                plugin_config,
                                registry=registry,
                                is_optional=arg.default_value == Undefined,
                            ),
                        )
                    )

                    if arg.default_value != Undefined:
                        kwdefaults.append(ast.Constant(value=arg.default_value))

            body = []

            if value.description:
                body.append(ast.Expr(value=ast.Constant(value=value.description)))

            body.append(ast.Return(value=ast.Constant(value=None)))

            returns = generate_object_field_annotation(
                value.type,
                classname,
                config,
                plugin_config,
                registry,
                is_optional=True,
            )

            keywords += generate_directive_keywords(value.ast_node, plugin_config)

            if not additional_args and key not in ["Mutation", "Subscription", "Query"]:
                if not keywords:
                    assign_value = None
                else:
                    assign_value = ast.Call(
                        func=ast.Name(
                            id="strawberry.field",
                            ctx=ast.Load(),
                        ),
                        keywords=keywords,
                        args=[],
                    )

                # lets create a simple field
                assign = ast.AnnAssign(
                    target=ast.Name(
                        registry.generate_node_name(value_key), ctx=ast.Store()
                    ),
                    annotation=returns,
                    value=assign_value,
                    simple=1,
                )

            else:
                if key == "Mutation":
                    decorator_name = ast.Name(id="strawberry.mutation", ctx=ast.Load())
                    function_def = ast.FunctionDef
                    returns = returns
                elif key == "Subscription":
                    registry.register_import("typing.AsyncGenerator")
                    decorator_name = ast.Name(
                        id="strawberry.subscription", ctx=ast.Load()
                    )
                    function_def = ast.AsyncFunctionDef
                    returns = ast.Subscript(
                        value=ast.Name(id="AsyncGenerator", ctx=ast.Load()),
                        slice=ast.Tuple(
                            elts=[returns, ast.Name(id="None", ctx=ast.Load())]
                        ),
                        ctx=ast.Load(),
                    )
                else:
                    decorator_name = ast.Name(id="strawberry.field", ctx=ast.Load())
                    function_def = ast.FunctionDef
                    returns = returns

                field_decorator = ast.Call(
                    func=decorator_name,
                    keywords=keywords,
                    args=[],
                )

                assign = function_def(
                    name=registry.generate_node_name(value_key),
                    returns=returns,
                    args=ast.arguments(
                        args=[ast.arg(arg="self")] + additional_args,
                        posonlyargs=[],
                        kwonlyargs=[],
                        kw_defaults=[],
                        defaults=kwdefaults,
                    ),
                    body=body,
                    decorator_list=[field_decorator],
                    simple=1,
                )

            fields += [assign]

        tree.append(
            ast.ClassDef(
                classname,
                bases=additional_bases,
                decorator_list=[decorator],
                keywords=[],
                body=fields
                + generate_pydantic_config(GraphQLTypes.OBJECT, config, registry, key),
            )
        )

    return tree


def generate_scalars(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: StrawberryPluginConfig,
    registry: ClassRegistry,
):
    objects = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLScalarType)
        and key not in plugin_config.builtin_scalars
    }

    tree = []

    if objects:
        registry.register_import("typing.NewType")

    for key, scalar in objects.items():
        keywords = []
        keywords += generate_directive_keywords(scalar.ast_node, plugin_config)
        if scalar.description:
            keywords.append(
                ast.keyword(
                    arg="description",
                    value=ast.Constant(value=scalar.description),
                )
            )

        keywords.append(
            ast.keyword(
                arg="serialize",
                value=ast.Lambda(
                    args=[
                        ast.arg(
                            arg="value",
                        ),
                    ],
                    body=ast.Name(id="value", ctx=ast.Load()),
                ),
            )
        )

        keywords.append(
            ast.keyword(
                arg="parse_value",
                value=ast.Lambda(
                    args=[
                        ast.arg(
                            arg="value",
                        ),
                    ],
                    body=ast.Name(id="value", ctx=ast.Load()),
                ),
            )
        )

        tree.append(
            ast.Assign(
                targets=[ast.Name(id=key, ctx=ast.Store())],
                value=ast.Call(
                    func=ast.Name(
                        id="strawberry.scalar",
                        ctx=ast.Load(),
                    ),
                    keywords=keywords,
                    args=[
                        ast.Call(
                            func=ast.Name(id="NewType", ctx=ast.Load()),
                            args=[
                                ast.Constant(value=key),
                                ast.Name(id="str", ctx=ast.Load()),
                            ],
                            keywords=[],
                        ),
                    ],
                ),
            )
        )
        registry.register_scalar(key, key)

    return tree


class StrawberryPlugin(Plugin):
    """Generate Strawberry models for GraphQL objects

    This plugin generates Pydantic models for GraphQL ObjectTypes and Interfaces in your graphql schema.

    Attention: This plugin is not made for client side usage. Most likely you are looking for the
    operations, fragments and funcs plugin (that generates client side validation and query function
    for your documents (your operations, fragments and queries).


    """

    config: StrawberryPluginConfig = Field(default_factory=StrawberryPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:
        registry.register_import("strawberry")

        directives = (
            self.config.generate_directives_func(
                client_schema, config, self.config, registry
            )
            if self.config.generate_directives
            else []
        )

        scalars = (
            generate_scalars(client_schema, config, self.config, registry)
            if self.config.generate_directives
            else []
        )

        enums = (
            self.config.generate_enums_func(
                client_schema, config, self.config, registry
            )
            if self.config.generate_enums
            else []
        )
        inputs = (
            generate_inputs(client_schema, config, self.config, registry)
            if self.config.generate_inputs
            else []
        )
        types = (
            generate_types(client_schema, config, self.config, registry)
            if self.config.generate_types
            else []
        )

        return directives + scalars + enums + inputs + types



from pydantic import Field
from turms.processors.base import Processor, ProcessorConfig
import libcst as cst
from collections import OrderedDict
from turms.config import GeneratorConfig
import os

# load beasts.py as a ast.Module


def retrieve_symbol_name(simple_statement: cst.SimpleStatementLine):
    for node in simple_statement.body:
        if isinstance(node, cst.AnnAssign):
            return node.target.value
    else:
        return None


def merge_functions(generated: cst.FunctionDef, existing: cst.FunctionDef):
    # merge the two functions

    # we want to merge the body of the new function into the existing one

    new_def = generated.with_changes(body=existing.body)
    return new_def


def merge_statements(
    generated: cst.SimpleStatementLine, existing: cst.SimpleStatementLine
):
    # merge the two statements

    # we want to merge the body of the new statement into the existing one

    new_def = existing.with_changes(body=generated.body)
    return new_def


def merge_class(generated: cst.ClassDef, existing: cst.ClassDef):
    # merge the two classes
    body_symbols = OrderedDict()
    body_implemented_symbols = []
    body_symbol_position = OrderedDict()

    generated_indented_block = generated.body

    for node in generated_indented_block.body:
        if isinstance(node, cst.FunctionDef):
            body_symbols[node.name.value] = node

        if isinstance(node, cst.SimpleStatementLine):
            x = retrieve_symbol_name(node)
            if retrieve_symbol_name(node) is not None:
                body_symbols[x] = node

    existing_indented_block = existing.body

    new_body = []

    for node in existing_indented_block.body:
        if isinstance(node, cst.SimpleStatementLine):
            name = retrieve_symbol_name(node)
            if name is not None:
                if name in body_symbols:
                    body_symbol_position[name] = len(new_body)
                    new_body.append(merge_statements(body_symbols[name], node))
                    body_implemented_symbols.append(name)
                else:
                    new_body.append(node)
            else:
                new_body.append(node)

        elif isinstance(node, cst.FunctionDef):
            if node.name.value in body_symbols:
                x = body_symbols[node.name.value]
                body_symbol_position[node.name.value] = len(new_body)
                new_body.append(merge_functions(x, node))
                body_implemented_symbols.append(node.name.value)
            else:
                new_body.append(node)

        else:
            new_body.append(node)

    # add the missing classes at the right position

    body_missing_symbols = [
        key for key in body_symbols.keys() if key not in body_implemented_symbols
    ]

    beforemap = {}  # a map index and a list of symbols to insert before
    aftermap = {}

    # add the missing classes at the right position
    last_key = None
    for existingkey in body_implemented_symbols:
        last_key = existingkey
        for key, value in body_symbols.items():
            if existingkey == key:
                break
            if key in body_missing_symbols:
                body_missing_symbols.remove(key)
                beforemap.setdefault(body_symbol_position[existingkey], []).append(
                    key
                )  # we need to reverse

    if last_key is not None:
        aftermap[body_symbol_position[last_key]] = body_missing_symbols

    updated_body = []
    for index, node in enumerate(new_body):
        if index in beforemap:
            for missingkey in beforemap[index]:
                updated_body.append(body_symbols[missingkey])
        updated_body.append(node)
        if index in aftermap:
            for missingkey in aftermap[index]:
                updated_body.append(body_symbols[missingkey])

    if last_key is None:
        for missingkey in body_missing_symbols:
            updated_body.append(body_symbols[missingkey])

    new_class = existing.with_changes(
        body=existing_indented_block.with_changes(body=updated_body)
    )

    return new_class


class MergeProcessorConfig(ProcessorConfig):
    type: str = "turms.processors.merge.MergeProcessor"


def merge_code(old_code: str, new_code: str, config: MergeProcessorConfig):

    existing_module = cst.parse_module(old_code)
    new_module = cst.parse_module(new_code)

    symbols = OrderedDict()

    implemented_symbols = []
    new_symbols = set()

    new_body = []

    # merge the two ast.Module
    for node in new_module.body:

        if isinstance(node, cst.ClassDef):
            symbols[node.name.value] = node
            new_symbols.add(node.name.value)

        if isinstance(node, cst.FunctionDef):
            symbols[node.name.value] = node
            new_symbols.add(node.name.value)

    symbol_position = {}

    for node in existing_module.body:
        if isinstance(node, cst.ClassDef):
            if node.name.value in symbols:
                x = symbols[node.name.value]
                # merge the two classes

                symbol_position[node.name.value] = len(new_body)
                new_body.append(merge_class(x, node))
                implemented_symbols.append(node.name.value)
            else:
                new_body.append(node)

        elif isinstance(node, cst.FunctionDef):
            if node.name.value in symbols:
                # merge the two functions
                x = symbols[node.name.value]

                symbol_position[node.name.value] = len(new_body)
                new_body.append(merge_functions(x, node))
                implemented_symbols.append(node.name.value)
            else:
                new_body.append(node)

        else:
            new_body.append(node)

    # add the missing classes at the right position

    missing_symbols = [key for key in symbols.keys() if key not in implemented_symbols]

    beforemap = {}  # a map index and a list of symbols to insert before
    aftermap = {}

    # add the missing classes at the right position
    last_key = None
    for existingkey in implemented_symbols:
        last_key = existingkey
        for key, value in symbols.items():
            if existingkey == key:
                break
            if key in missing_symbols:
                missing_symbols.remove(key)
                beforemap.setdefault(symbol_position[existingkey], []).append(
                    key
                )  # we need to reverse

    if last_key is not None:
        aftermap[symbol_position[last_key]] = missing_symbols

    updated_body = []
    for index, node in enumerate(new_body):
        if index in beforemap:
            for missingkey in beforemap[index]:
                updated_body.append(symbols[missingkey])
        updated_body.append(node)
        if index in aftermap:
            for missingkey in aftermap[index]:
                updated_body.append(symbols[missingkey])

    if last_key is None:
        for missingkey in missing_symbols:
            updated_body.append(symbols[missingkey])

    md = cst.Module(body=updated_body)

    return md.code


class MergeProcessor(Processor):
    """A processor that merges the newly generated code
    with the existing code in the file.

    This uiliizes a third party library called libcst to
    parse the existing code and the newly generated code
    and then merge them together.

    We need to use libcst because ast.parse does not
    preserve comments and formatting.

    Important:
        This processor is experimental and may not work
        as expected. Please report any issues you find
        on the github repo.

    """

    config: MergeProcessorConfig = Field(default_factory=MergeProcessorConfig)

    def run(self, gen_file: str, config: GeneratorConfig):

        old_generated_file = os.path.join(
            config.out_dir,
            config.generated_name,
        )

        if not os.path.exists(old_generated_file):
            print("No existing generated file found. Skipping merge.")
            return gen_file

        # the one we want to merge into
        with open(old_generated_file) as f:
            existing_code = f.read()

        return merge_code(existing_code, gen_file, self.config)



from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.language.ast import (
    FieldNode,
    FragmentSpreadNode,
    InlineFragmentNode,
)
from turms.registry import ClassRegistry
from turms.utils import (
    generate_generic_typename_field,
    generate_pydantic_config,
    generate_typename_field,
    get_additional_bases_for_type,
    get_interface_bases,
    non_typename_fields,
    target_from_node,
)
import ast
from graphql.type.definition import (
    GraphQLEnumType,
    GraphQLField,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLType,
    GraphQLUnionType,
)

from turms.config import GraphQLTypes


def recurse_annotation(
    node: FieldNode,
    parent: str,
    type: GraphQLType,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    subtree: ast.AST,
    registry: ClassRegistry,
    is_optional=True,
) -> ast.AST:
    """Recurse Annotations

    Resolves the type of a field and returns the appropriate annotation.
    If we deal with nested object types it recurses further and generated the objects
    together with the type_field_node method:

    class X(BaseModel):
        a: int

    in this case "a" is generated by type_field_node and "X" is generated by recurse_annotation


    Args:
        node (FieldNode): The node
        type (GraphQLType): The type of the field as specified in the schema
        client_schema (GraphQLSchema): The schema itself
        config (GeneratorConfig): The generator config (with the defaults)
        subtree (ast.AST): The passed subtree
        registry (ClassRegistry): A class registry where classes and their imports are registered
        parent_name (str, optional): If resolving nested types the name of parent. Defaults to "".
        is_optional (bool, optional): A recurse modifier for optional types. Defaults to True.

    Raises:
        NotImplementedError: Not implemneted errors
        NotImplementedError: _description_

    Returns:
        ast.AST: The returned tree
    """

    if isinstance(type, GraphQLUnionType):

        union_class_names = []

        sub_nodes = non_typename_fields(node)

        for sub_node in sub_nodes:

            if isinstance(sub_node, FragmentSpreadNode):
                fragment_name = registry.inherit_fragment(sub_node.name.value)
                union_class_names.append(fragment_name)

            if isinstance(sub_node, InlineFragmentNode):
                inline_fragment_name = (
                    f"{parent}{sub_node.type_condition.name.value}InlineFragment"
                )
                inline_fragment_fields = []

                inline_fragment_fields += [
                    generate_typename_field(
                        sub_node.type_condition.name.value, registry, config
                    )
                ]

                for sub_sub_node in sub_node.selection_set.selections:

                    if isinstance(sub_sub_node, FieldNode):
                        sub_sub_node_type = client_schema.get_type(
                            sub_node.type_condition.name.value
                        )

                        if sub_sub_node.name.value == "__typename":
                            continue

                        field_type = sub_sub_node_type.fields[sub_sub_node.name.value]
                        inline_fragment_fields += type_field_node(
                            sub_sub_node,
                            inline_fragment_name,
                            field_type,
                            client_schema,
                            config,
                            subtree,
                            registry,
                        )

                additional_bases = get_additional_bases_for_type(
                    sub_node.type_condition.name.value, config, registry
                )
                cls = ast.ClassDef(
                    inline_fragment_name,
                    bases=additional_bases
                    + [
                        ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                        for base in config.object_bases
                    ],
                    decorator_list=[],
                    keywords=[],
                    body=inline_fragment_fields
                    + generate_pydantic_config(GraphQLTypes.FRAGMENT, config, registry),
                )

                subtree.append(cls)
                union_class_names.append(inline_fragment_name)

        assert (
            len(union_class_names) != 0
        ), f"You have set 'always_resolve_interfaces' to True but you have no sub-fragments in your query of {base_name}"

        if len(union_class_names) > 1:
            registry.register_import("typing.Union")
            union_slice = ast.Tuple(
                elts=[
                    ast.Name(id=clsname, ctx=ast.Load())
                    for clsname in union_class_names
                ],
                ctx=ast.Load(),
            )

            if is_optional:
                registry.register_import("typing.Optional")

                return ast.Subscript(
                    value=ast.Name("Optional", ctx=ast.Load()),
                    slice=ast.Subscript(
                        value=ast.Name("Union", ctx=ast.Load()),
                        slice=union_slice,
                    ),
                    ctx=ast.Load(),
                )
            else:
                registry.register_import("typing.Union")
                return ast.Subscript(
                    value=ast.Name("Union", ctx=ast.Load()),
                    slice=union_slice,
                    ctx=ast.Load(),
                )
        else:
            return ast.Name(id=union_class_names[0], ctx=ast.Load())

    if isinstance(type, GraphQLInterfaceType):
        # Lets Create Base Class to Inherit from for this
        mother_class_fields = []
        target = target_from_node(node)

        sub_nodes = non_typename_fields(node)

        base_name = f"{parent}{target.capitalize()}"

        if type.description:
            mother_class_fields.append(
                ast.Expr(value=ast.Constant(value=type.description))
            )

        implementing_types = client_schema.get_implementations(type)


        implementing_class_base_classes = {

        }

        inline_fragment_fields = {}


        

        for sub_node in sub_nodes:

            if isinstance(sub_node, FragmentSpreadNode):
                # Spread nodes are like inheritance?
                try:
                    # We are dealing with a fragment that is an interface
                    implementation_map = registry.get_interface_fragment_implementations(sub_node.name.value)
                    for k, v in implementation_map.items():
                        implementing_class_base_classes.setdefault(k, []).append(v)

                except KeyError:
                    x = registry.get_fragment_type(sub_node.name.value)
                    implementing_class_base_classes.setdefault(x.name, []).append(registry.inherit_fragment(sub_node.name.value))


            if isinstance(sub_node, FieldNode):

                field_type = type.fields[sub_node.name.value]
                mother_class_fields += type_field_node(
                    sub_node,
                    base_name,
                    field_type,
                    client_schema,
                    config,
                    subtree,
                    registry,
                )

            if isinstance(sub_node, InlineFragmentNode):


                on_type_name = sub_node.type_condition.name.value

                inline_fragment_fields.setdefault(on_type_name, []).append(
                    generate_typename_field(
                        sub_node.type_condition.name.value, registry, config
                    )
                )





        # We first genrate the mother class that will provide common fields of this fragment. This will never be reference
        # though
        mother_class_name = f"{base_name}Base"
        additional_bases = get_additional_bases_for_type(type.name, config, registry)

        body = mother_class_fields if mother_class_fields else [ast.Pass()]



        mother_class = ast.ClassDef(
            mother_class_name,
            bases=additional_bases + get_interface_bases(config, registry),
            decorator_list=[],
            keywords=[],
            body=body
            + generate_pydantic_config(GraphQLTypes.FRAGMENT, config, registry),
        )
        subtree.append(mother_class)

        implementaionMap = {}
        union_class_names = []

        for i in implementing_types.objects:

            class_name = f"{mother_class_name}{i.name}"

            ast_base_nodes = [ast.Name(id=x, ctx=ast.Load()) for x in implementing_class_base_classes.get(i.name, [])]
            implementaionMap[i.name] = class_name

            inline_fields = inline_fragment_fields.get(i, [])
        

            implementing_class = ast.ClassDef(
                    class_name,
                    bases=ast_base_nodes + [ast.Name(id=mother_class_name, ctx=ast.Load())] + get_interface_bases(config, registry),  # Todo: fill with base
                    decorator_list=[],
                    keywords=[],
                    body=[generate_typename_field(i.name, registry, config)] + inline_fields,
            )

            subtree.append(implementing_class)
            union_class_names.append(class_name)



        registry.register_import("typing.Annotated")
        registry.register_import("typing.Union")
        union_slice = ast.Tuple(
            elts=[
                ast.Name(id=clsname, ctx=ast.Load())
                for clsname in union_class_names
            ],
            ctx=ast.Load(),
        )

        slice = ast.Tuple(
            elts=[
                ast.Subscript(
                    value=ast.Name("Union", ctx=ast.Load()),
                    slice=union_slice,
                    ctx=ast.Load(),
                ),
                ast.Call(
                    func=ast.Name(id="Field", ctx=ast.Load()),
                    args=[],
                    keywords=[ast.keyword(arg="discriminator", value=ast.Constant("typename"))],
                )
            ],
            ctx=ast.Load(),
        )

        # Resort to base class if we have no sub-fragments
        annotated_slice = ast.Subscript(
                    value=ast.Name("Annotated", ctx=ast.Load()),
                    slice=slice,
                    ctx=ast.Load(),
                )
        





        if is_optional:
            registry.register_import("typing.Optional")

            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=annotated_slice,
                ctx=ast.Load(),
            )
        else:
            registry.register_import("typing.Union")
            return annotated_slice

    if isinstance(type, GraphQLObjectType):
        pick_fields = []

        target = target_from_node(node)
        object_class_name = f"{parent}{target.capitalize()}"

        if type.description:
            pick_fields.append(ast.Expr(value=ast.Constant(value=type.description)))

        pick_fields += [generate_typename_field(type.name, registry, config)]

        sub_nodes = non_typename_fields(node)

        # Single Item collapse
        if len(sub_nodes) == 1:
            sub_node = sub_nodes[0]

            if isinstance(sub_node, FragmentSpreadNode):

                if is_optional:
                    registry.register_import("typing.Optional")
                    return ast.Subscript(
                        value=ast.Name("Optional", ctx=ast.Load()),
                        slice=registry.reference_fragment(
                            sub_node.name.value, parent
                        ),  # needs to be parent not object as reference will be to parent
                        ctx=ast.Load(),
                    )

                else:
                    return registry.reference_fragment(
                        sub_node.name.value, parent
                    )  # needs to be parent not object as reference will be to parent



        additional_bases = []

        for sub_node in sub_nodes:

            if isinstance(sub_node, FragmentSpreadNode):

                if registry.is_interface_fragment(sub_node.name.value):
                    raise Exception("Interface Fragments with additional subfields are not yet implemented")
                
                else:
                    additional_bases.append(
                        ast.Name(
                            id=registry.inherit_fragment(sub_node.name.value),
                            ctx=ast.Load(),
                        )
                    )

            if isinstance(sub_node, FieldNode):
                if sub_node.name.value == "__typename":
                    continue
                field_type = type.fields[sub_node.name.value]
                pick_fields += type_field_node(
                    sub_node,
                    object_class_name,
                    field_type,
                    client_schema,
                    config,
                    subtree,
                    registry,
                )

            if isinstance(sub_node, InlineFragmentNode):
                raise NotImplementedError("Inline Fragments are not yet implemented")
            
        if not additional_bases:
            # We need to add the base class if we have no fragments
            additional_bases = get_additional_bases_for_type(type.name, config, registry)
            

        body = pick_fields if pick_fields else [ast.Pass()]

        cls = ast.ClassDef(
            object_class_name,
            bases=additional_bases
            + [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.object_bases
            ],
            decorator_list=[],
            keywords=[],
            body=body + generate_pydantic_config(GraphQLTypes.OBJECT, config, registry),
        )

        subtree.append(cls)

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Name(
                    id=object_class_name,
                    ctx=ast.Load(),
                ),
                ctx=ast.Load(),
            )

        else:
            return ast.Name(
                id=object_class_name,
                ctx=ast.Load(),
            )

    if isinstance(type, GraphQLScalarType):

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(type.name),
            )

        else:
            return registry.reference_scalar(type.name)

    if isinstance(type, GraphQLEnumType):

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_enum(type.name, parent),
            )

        else:
            return registry.reference_enum(type.name, parent)

    if isinstance(type, GraphQLNonNull):
        return recurse_annotation(
            node,
            parent,
            type.of_type,
            client_schema,
            config,
            subtree,
            registry,
            is_optional=False,
        )

    if isinstance(type, GraphQLList):

        if config.freeze.enabled:
            registry.register_import("typing.Tuple")

            def list_builder(x):
                return ast.Subscript(
                    value=ast.Name("Tuple", ctx=ast.Load()),
                    slice=ast.Tuple(elts=[x, ast.Constant(value=...)], ctx=ast.Load()),
                    ctx=ast.Load(),
                )

        else:

            registry.register_import("typing.List")

            def list_builder(x):
                return ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()), slice=x, ctx=ast.Load()
                )

        if is_optional:
            registry.register_import("typing.Optional")

            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=list_builder(
                    recurse_annotation(
                        node,
                        parent,
                        type.of_type,
                        client_schema,
                        config,
                        subtree,
                        registry,
                    )
                ),
                ctx=ast.Load(),
            )

        else:
            return list_builder(
                recurse_annotation(
                    node,
                    parent,
                    type.of_type,
                    client_schema,
                    config,
                    subtree,
                    registry,
                )
            )


def type_field_node(
    node: FieldNode,
    parent: str,
    field: GraphQLField,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    subtree: ast.AST,
    registry: ClassRegistry,
    is_optional=True,
):
    """Types a field node

    This

    Args:
        node (FieldNode): _description_
        field (GraphQLField): _description_
        client_schema (GraphQLSchema): _description_
        config (GeneratorConfig): _description_
        subtree (ast.AST): _description_
        registry (ClassRegistry): _description_
        parent_name (str, optional): _description_. Defaults to "".
        is_optional (bool, optional): _description_. Defaults to True.

    Returns:
        _type_: _description_
    """

    target = target_from_node(node)
    field_name = registry.generate_node_name(target)

    keywords = []
    if not isinstance(field.type, GraphQLNonNull):
        keywords.append(ast.keyword(arg="default", value=ast.Constant(value=None)))

    if target != field_name:
        registry.register_import("pydantic.Field")
        assign = ast.AnnAssign(
            target=ast.Name(field_name, ctx=ast.Store()),
            annotation=recurse_annotation(
                node,
                parent,
                field.type,
                client_schema,
                config,
                subtree,
                registry,
                is_optional=is_optional,
            ),
            value=ast.Call(
                func=ast.Name(id="Field", ctx=ast.Load()),
                args=[],
                keywords=keywords
                + [ast.keyword(arg="alias", value=ast.Constant(value=target))],
            ),
            simple=1,
        )
    else:
        if keywords:
            registry.register_import("pydantic.Field")
        assign = ast.AnnAssign(
            target=ast.Name(target, ctx=ast.Store()),
            annotation=recurse_annotation(
                node,
                parent,
                field.type,
                client_schema,
                config,
                subtree,
                registry,
                is_optional=is_optional,
            ),
            value=(
                ast.Call(
                    func=ast.Name(id="Field", ctx=ast.Load()),
                    args=[],
                    keywords=keywords,
                )
                if keywords
                else None
            ),
            simple=1,
        )

    potential_comment = (
        field.description
        if not field.deprecation_reason
        else f"DEPRECATED {field.deprecation_reason}: : {field.description} "
    )

    if field.deprecation_reason:
        registry.warn(
            f"Field {node.name.value} on {parent} is deprecated: {field.deprecation_reason}"
        )

    if potential_comment:
        return [assign, ast.Expr(value=ast.Constant(value=potential_comment))]

    return [assign]



import builtins
from pydantic import (
    AnyHttpUrl,
    BaseModel,
    Field,
    GetCoreSchemaHandler,
    field_validator,
    validator,
    ConfigDict,
)
from pydantic_core import core_schema
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Union,
    Protocol,
    Literal,
    runtime_checkable,
)
from turms.helpers import import_string
from enum import Enum


class ConfigProxy(BaseModel):
    model_config = ConfigDict(extra="allow")
    type: str


class ImportableFunctionMixin(Protocol):

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls.validate, handler(callable), field_name=handler.field_name
        )

    @classmethod
    def validate(cls, v, *info):
        if not callable(v):
            if not isinstance(v, str):
                raise TypeError("string required")
            assert "." in v, "You need to point to a module if its not a builtin type"
            v = import_string(v)

        assert callable(v)
        return v


class PythonType(str):
    """A string that represents a python type. Either a builtin type or a type from a module."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls.validate, handler(str), field_name=handler.field_name
        )

    @classmethod
    def validate(cls, v, *info):
        if not isinstance(v, str):
            raise TypeError("string required")
        if v not in dir(builtins):
            assert "." in v, "You need to point to a module if its not a builtin type"
        return cls(v)


class GraphQLTypes(str, Enum):
    INPUT: str = "input"
    FRAGMENT: str = "fragment"
    OBJECT: str = "object"
    MUTATION = "mutation"
    QUERY = "query"
    SUBSCRIPTION = "subscription"
    DIRECTIVE: str = "directive"


class LogLevel(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


@runtime_checkable
class LogFunction(Protocol):
    def __call__(self, message, level: LogLevel = LogLevel.INFO):
        pass


class FreezeConfig(BaseSettings):
    """Configuration for freezing the generated pydantic
    models

    This is useful for when you want to generate the models
    that are faux immutable and hashable by default. The configuration
    allows you to customize the way the models are frozen and specify
    which types (operation, fragment, input,...) should be frozen.

    """

    enabled: bool = Field(False, description="Enabling this, will freeze the schema")
    """Enabling this, will freeze the schema"""

    types: List[GraphQLTypes] = Field(
        [GraphQLTypes.INPUT, GraphQLTypes.FRAGMENT, GraphQLTypes.OBJECT],
        description="The types to freeze",
    )
    """The core types (Input, Fragment, Object, Operation) to freeze"""

    exclude: Optional[List[str]] = Field(
        None, description="List of types to exclude from freezing"
    )
    """List of types to exclude from freezing"""
    include: Optional[List[str]] = Field(
        None, description="List of types to include in freezing"
    )
    """The types to freeze"""
    exclude_fields: Optional[List[str]] = Field(
        [], description="List of fields to exclude from freezing"
    )
    include_fields: Optional[List[str]] = Field(
        [], description="List of fields to include in freezing"
    )
    convert_list_to_tuple: bool = Field(
        True, description="Convert GraphQL List to tuple (with varying length"
    )
    """Convert GraphQL List to tuple (with varying length)"""


ExtraOptions = Optional[Union[Literal["ignore"], Literal["allow"], Literal["forbid"]]]


class OptionsConfig(BaseSettings):
    """Configuration for freezing the generated pydantic
    models

    This is useful for when you want to generate the models
    that are faux immutable and hashable by default. The configuration
    allows you to customize the way the models are frozen and specify
    which types (operation, fragment, input,...) should be frozen.

    """

    enabled: bool = Field(False, description="Enabling this, will freeze the schema")
    """Enabling this, will freeze the schema"""
    extra: ExtraOptions = None
    """Extra options for pydantic"""
    allow_mutation: Optional[bool] = None
    """Allow mutation"""
    allow_population_by_field_name: Optional[bool] = None
    """Allow population by field name"""
    orm_mode: Optional[bool] = None
    """ORM mode"""
    use_enum_values: Optional[bool] = None
    """Use enum values"""

    validate_assignment: Optional[bool] = None
    """Validate assignment"""

    types: List[GraphQLTypes] = Field(
        [GraphQLTypes.INPUT, GraphQLTypes.FRAGMENT, GraphQLTypes.OBJECT],
        description="The types to freeze",
    )
    """The core types (Input, Fragment, Object, Operation) to enable this option"""

    exclude: Optional[List[str]] = Field(
        None, description="List of types to exclude from setting this option"
    )
    """List of types to exclude from setting this option"""
    include: Optional[List[str]] = Field(
        None, description="List of types to include in setting these options"
    )
    """The types to freeze"""


PydanticVersion = Literal["v1", "v2"]


class GeneratorConfig(BaseSettings):
    """Configuration for the generator

    This is the main generator configuration that allows you to
    customize the way the models are generated.

    You need to specify the documents that should be parsed
    and the scalars that should be used.

    """

    model_config: SettingsConfigDict = SettingsConfigDict(
        env_prefix="TURMS_",
        extra="forbid",
    )
    pydantic_version: PydanticVersion = "v2"

    domain: Optional[str] = None
    """The domain of the GraphQL API ( will be set as a config variable)"""
    out_dir: str = "api"
    """The output directory for the generated models"""
    dump_configuration: bool = False
    configuration_name: str = "project.json"
    dump_schema: bool = False
    schema_name: str = "schema.graphql"
    generated_name: str = "schema.py"
    """ The name of the generated file within the output directory"""
    documents: Optional[str] = None
    """The documents to parse. Setting this will overwrite the documents in the graphql config"""
    verbose: bool = False
    """Enable verbose logging"""

    exit_on_error: bool = True
    """Will cause a sys.exit(1) if an error occurs"""

    allow_introspection: bool = True
    """Allow introspection queries"""

    object_bases: List[str] = ["pydantic.BaseModel"]
    """The base classes for the generated objects. This is useful if you want to change the base class from BaseModel to something else"""

    interface_bases: Optional[List[str]] = None
    """List of base classes for interfaces"""
    always_resolve_interfaces: bool = True
    """Always resolve interfaces to concrete types"""
    exclude_typenames: bool = False
    """Exclude __typename from generated models when calling dict or json"""

    scalar_definitions: Dict[str, PythonType] = Field(
        default_factory=dict,
        description="Additional config for mapping scalars to python types (e.g. ID: str). Can use dotted paths to import types from other modules.",
    )
    """Additional config for mapping scalars to python types (e.g. ID: str). Can use dotted paths to import types from other modules."""
    freeze: FreezeConfig = Field(
        default_factory=FreezeConfig,
        description="Configuration for freezing the generated models",
    )
    """Configuration for freezing the generated models: by default disabled"""

    options: OptionsConfig = Field(
        default_factory=OptionsConfig,
        description="Configuration for pydantic options",
    )
    """Configuration for pydantic options: by default disabled"""

    skip_forwards: bool = False
    """Skip generating automatic forwards reference for the generated models"""

    additional_bases: Dict[str, List[str]] = Field(
        default_factory=dict,
        description="Additional bases for the generated models as map of GraphQL Type to importable base class (e.g. module.package.Class)",
    )
    "Additional bases for the generated models as map of GraphQL Type to importable base class (e.g. module.package.Class)"
    additional_config: Dict[str, Dict[str, Any]] = Field(
        default_factory=dict,
        description="Additional config for the generated models as map of GraphQL Type to config attributes",
    )
    "Additional config for the generated models as map of GraphQL Type to config attributes"

    force_plugin_order: bool = True
    "Should the plugins be forced to run in the order they are defined"

    parsers: List[ConfigProxy] = Field(
        default_factory=list,
        description="List of parsers to use. Parsers are used to parse the generated AST and translate it before it is converted to python code",
    )
    "List of parsers to use. Parsers are used to parse the generated AST and translate it before it is converted to python code"

    plugins: List[ConfigProxy] = Field(
        default_factory=list,
        description="List of plugins to use. Plugins are used to generated the python ast from the graphql documents, objects, etc.",
    )
    "List of plugins to use. Plugins are used to generated the python ast from the graphql documents, objects, etc."
    processors: List[ConfigProxy] = Field(
        default_factory=list,
        description="List of processors to use. Processor are used to enforce specific styles on the generated python code",
    )
    "List of processors to use. Processor are used to enforce specific styles on the generated python code"
    stylers: List[ConfigProxy] = Field(
        default_factory=list,
        description="List of stylers to use. Style are used to enforce specific styles on the generaded class or fieldnames. ",
    )
    "List of stylers to use. Style are used to enforce specific styles on the generaded class or fieldnames. "

    @field_validator("parsers", "plugins", "processors", "stylers")
    def validate_importable(cls, v):
        try:
            for parser in v:
                import_string(parser.type)
        except Exception as e:
            raise ValueError(f"Invalid import: {parser.type} {e}") from e

        return v
    

    @field_validator("additional_bases")
    def validate_additional_bases(cls, v):
        for key, value_list in v.items():
            for value in value_list:
                if not isinstance(value, str):
                    raise ValueError("string required")
                if value not in dir(builtins):
                    if "." not in value:
                        raise ValueError("You need to point to a module if its not a builtin type")
        return v



class Extensions(BaseModel):
    """Wrapping class to be able to extract the tums configuraiton"""

    turms: GeneratorConfig
    "The turms configuration"


class AdvancedSchemaField(BaseModel):
    headers: Dict[str, str]


SchemaField = Union[AnyHttpUrl, str, Dict[str, AdvancedSchemaField]]
SchemaType = Union[SchemaField, List[SchemaField]]


class GraphQLProject(BaseSettings):
    """Configuration for the GraphQL project

    This is the main configuration for one GraphQL project. It is compliant with
    the graphql-config specification. And allows you to specify the schema and
    the documents that should be parsed.

    Turm will use the schema and documents to generate the python models, according
    to the generator configuration under extensions.turms
    """

    model_config: SettingsConfigDict = SettingsConfigDict(
        env_prefix="TURMS_GRAPHQL_",
        extra="allow",
    )

    schema_url: SchemaType = Field(alias="schema")
    """The schema url or path to the schema file"""
    documents: Optional[str] = None
    """The documents (operations,fragments) to parse"""
    extensions: Extensions
    """The extensions configuration for the project (here resides the turms configuration)"""


class GraphQLConfigMultiple(BaseSettings):
    """Configuration for multiple GraphQL projects

    This is the main configuration for multiple GraphQL projects. It is compliant with
    the graphql-config specification for multiple projec."""

    model_config: SettingsConfigDict = SettingsConfigDict(
        extra="allow",
    )

    projects: Dict[str, GraphQLProject]
    """ The projects that should be parsed. The key is the name of the project and the value is the graphql project"""


class GraphQLConfigSingle(GraphQLProject):
    """Configuration for a single GraphQL project

    This is the main configuration for a single GraphQL project. It is compliant with
    the graphql-config specification for a single project.
    """

    model_config: SettingsConfigDict = SettingsConfigDict(
        extra="allow",
    )



class TurmsError(Exception):
    """Base class for all Turms errors"""

    pass


class RegistryError(Exception):
    """Base class for all registry errors"""

    pass


class NoScalarFound(RegistryError):
    """Raised when no scalar is found for a given type. Needs to provided by additional scalars"""

    pass


class NoInputTypeFound(RegistryError):
    """Raised when no input type is found for a given type. Often raised if the input plugin is not loaded, and operations
    use input tyes"""

    pass


class NoEnumFound(RegistryError):
    """Raised when no enum is found for a given type. Often raised if the enum plugin is not loaded, and operations or fragments
    use enums"""

    pass


class GenerationError(TurmsError):
    """Base class for all generation errors"""

    pass



import ast
from typing import List
from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class ParserConfig(BaseSettings):
    model_config = SettingsConfigDict(extra="forbid")
    type: str


class Parser(BaseModel):
    """Base class for all parsers

    Parsers are used to parse the AST of the generated python code. They can be used to
    modify the AST before it is written to the file."""

    config: ParserConfig

    def parse_ast(
        self,
        asts: List[ast.AST],
    ) -> List[ast.AST]: ...  # pragma: no cover



from __future__ import annotations

import ast
import logging
import re
from typing import Any, List, Optional, Tuple


from graphql import (
    FragmentSpreadNode,
    GraphQLInputObjectType,
    GraphQLObjectType,
    Undefined,
    VariableNode,
    is_wrapping_type,
)
from graphql.type.definition import (
    GraphQLEnumType,
)
from graphql.language.ast import (
    FieldNode,
    NonNullTypeNode,
    OperationDefinitionNode,
    OperationType,
)
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.utilities.get_operation_root_type import get_operation_root_type
from graphql.utilities.type_info import get_field_def
from pydantic import BaseModel, Field
from pydantic_settings import SettingsConfigDict
from turms.config import GeneratorConfig
from turms.plugins.base import Plugin, PluginConfig
from turms.registry import ClassRegistry
from turms.utils import (
    inspect_operation_for_documentation,
    non_typename_fields,
    parse_documents,
    parse_value_node,
    recurse_outputtype_annotation,
    recurse_outputtype_label,
    recurse_type_annotation,
    recurse_type_label,
    target_from_node,
)
from graphql import (
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
)

logger = logging.getLogger(__name__)


class Kwarg(BaseModel):
    key: str
    type: str
    description: str = "Specify that in turms.plugin.funcs.OperationsFuncPlugin"
    default: Any = None


class Arg(BaseModel):
    key: str
    type: str
    description: str = "Specify that in turms.plugin.funcs.OperationsFuncPlugin"


class FunctionDefinition(BaseModel):
    type: OperationType
    is_async: bool = False
    extra_args: List[Arg] = []
    extra_kwargs: List[Kwarg] = []
    use: str


class FuncsPluginConfig(PluginConfig):
    model_config = SettingsConfigDict(extra="forbid", env_prefix="TURMS_PLUGINS_FUNCS_")
    type: str = "turms.plugins.funcs.FuncsPlugin"
    funcs_glob: Optional[str] = None
    prepend_sync: str = ""
    prepend_async: str = "a"
    collapse_lonely: bool = True
    generate_protocol: bool = False
    global_args: List[Arg] = []
    global_kwargs: List[Kwarg] = []
    definitions: List[FunctionDefinition] = []
    extract_documentation: bool = True
    argument_key_is_styled: bool = False
    expand_input_types: List[str] = []


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def generate_async_func_name(
    o: OperationDefinitionNode,
    plugin_config: FuncsPluginConfig,
    config: GeneratorConfig,
    registry: ClassRegistry,
):
    return f"{plugin_config.prepend_async}{camel_to_snake(o.name.value)}"


def generate_sync_func_name(
    o: OperationDefinitionNode,
    plugin_config: FuncsPluginConfig,
    config: GeneratorConfig,
    registry: ClassRegistry,
):
    return f"{plugin_config.prepend_sync}{camel_to_snake(o.name.value)}"


def get_extra_args_for_onode(
    definition: FunctionDefinition,
    plugin_config: FuncsPluginConfig,
) -> List[Arg]:
    args = plugin_config.global_args
    return args + definition.extra_args


def generate_passing_extra_args_for_onode(
    definition: FunctionDefinition, plugin_config: FuncsPluginConfig
):
    return [
        ast.Name(id=arg.key, ctx=ast.Load())
        for arg in get_extra_args_for_onode(definition, plugin_config)
    ]


def generate_passing_extra_kwargs_for_onode(
    definition: FunctionDefinition, plugin_config: FuncsPluginConfig
):
    return [
        ast.keyword(arg=kwarg.key, value=ast.Name(id=kwarg.key, ctx=ast.Load()))
        for kwarg in get_extra_kwargs_for_onode(definition, plugin_config)
    ]


def get_extra_kwargs_for_onode(
    definition: FunctionDefinition,
    plugin_config: FuncsPluginConfig,
) -> List[Kwarg]:
    kwargs = plugin_config.global_kwargs

    return kwargs + definition.extra_kwargs


def get_definitions_for_onode(
    operation_definition: OperationDefinitionNode,
    plugin_config: FuncsPluginConfig,
) -> List[Arg]:
    """Checks the Plugin Config if the operation definition should be included
    in the generated functions

    Args:
        operation_definition (OperationDefinitionNode): _description_
        plugin_config (FuncsPluginConfig): _description_

    Returns:
        List[Arg]: _description_
    """

    definitions = [
        definition
        for definition in plugin_config.definitions
        if definition.type == operation_definition.operation
    ]

    return definitions



def generate_input_annotation(
    type: GraphQLInputObjectType,
    registry: ClassRegistry,
    is_optional=True,
):
    if isinstance(type, GraphQLScalarType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_scalar(type.name),
            )

        return registry.reference_scalar(type.name)

    if isinstance(type, GraphQLInputObjectType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_inputtype(type.name, "SHOULD_NOT_BE_USED"),
                ctx=ast.Load(),
            )
        return registry.reference_inputtype(type.name, "SHOULD_NOT_BE_USED")

    if isinstance(type, GraphQLEnumType):
        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=registry.reference_enum(
                    type.name, "SHOULD_NOT_BE_USED", allow_forward=False
                ),
                ctx=ast.Load(),
            )
        return registry.reference_enum(
            type.name, "SHOULD_NOT_BE_USED", allow_forward=False
        )

    if isinstance(type, GraphQLNonNull):
        return generate_input_annotation(
            type.of_type, registry, is_optional=False
        )

    if isinstance(type, GraphQLList):
        registry.register_import("typing.Iterable")

        def list_builder(x):
            return ast.Subscript(
                value=ast.Name("Iterable", ctx=ast.Load()), slice=x, ctx=ast.Load()
            )

        if is_optional:
            registry.register_import("typing.Optional")
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=list_builder(
                    generate_input_annotation(
                        type.of_type,
                        registry,
                        is_optional=True,
                    )
                ),
                ctx=ast.Load(),
            )

        return list_builder(
            generate_input_annotation(
                type.of_type, registry, is_optional=True
            )
        )

    raise NotImplementedError(f"Unknown input type {type}")


def generate_input_description(
    type: GraphQLInputObjectType,
    registry: ClassRegistry,
    is_optional=True,
):
    if isinstance(type, GraphQLScalarType):
        return type.description or ""

    if isinstance(type, GraphQLInputObjectType):
        return type.description or ""

    if isinstance(type, GraphQLEnumType):
        return type.name or ""

    if isinstance(type, GraphQLNonNull):
        return generate_input_description(
            type.of_type, registry, is_optional=False
        ) + " (required)"
    
    if isinstance(type, GraphQLList):
        return generate_input_description(
            type.of_type, registry, is_optional=True
        ) + " (list)"

    raise NotImplementedError(f"Unknown input type {type}")


def generate_input_type_descriptions(input_type: GraphQLInputObjectType, registry: ClassRegistry):
    description = ""

    for value_key, value in input_type.fields.items():
        field_name = registry.generate_node_name(value_key)

    
        description += f"    {registry.generate_parameter_name(value_key)}: {value.description or generate_input_description(value.type, registry)}\n"

    return description



def generate_input_type_params(
    input_type: GraphQLInputObjectType, registry: ClassRegistry
):
    



    pos_args = []
    kw_args = []
    kw_values = []

    for value_key, value in input_type.fields.items():

        field_name = registry.generate_node_name(value_key)

        if isinstance(value.type, GraphQLNonNull):
            pos_args.append(
                ast.arg(
                    arg=field_name,
                    annotation=generate_input_annotation(
                        value.type,
                        registry,
                        is_optional=True,
                    ),
                )
            )
        else:
            kw_args.append(
                ast.arg(
                    arg=field_name,
                    annotation=generate_input_annotation(
                        value.type,
                        registry,
                        is_optional=True,
                    ),
                )
            )
            kw_values.append(ast.Constant(value=value.default_value if value.default_value and value.default_value is not Undefined else None))


    return pos_args, kw_args, kw_values





def generate_parameters(
    definition: FunctionDefinition,
    operation_definition: OperationDefinitionNode,
    config: GeneratorConfig,
    plugin_config: FuncsPluginConfig,
    registry: ClassRegistry,
    client_schema: GraphQLSchema,
):
    

    pos_args = []
    kw_args = []
    kw_values = []


    extra_args = get_extra_args_for_onode(definition, plugin_config)
    

    for arg in extra_args:
        registry.register_import(arg.type)
        pos_args.append(
            ast.arg(
                arg=arg.key,
                annotation=ast.Name(
                    id=arg.type.split(".")[-1],
                    ctx=ast.Load(),
                ),
            )
        )

    arg_variables = [
        v
        for v in operation_definition.variable_definitions
        if isinstance(v.type, NonNullTypeNode) and not v.default_value
    ]
    kwarg_variables = [
        v
        for v in operation_definition.variable_definitions
        if not isinstance(v.type, NonNullTypeNode) or v.default_value
    ]
    



    non_expandable_args = []

    for v in arg_variables:
        if v.variable.name.value in plugin_config.expand_input_types:

            input_type = v.type

            type = client_schema.type_map[input_type.type.name.value]



            expanded_pos_args, expanded_kw_args, expanded_kw_values = generate_input_type_params(
               type, registry
            )

            pos_args += expanded_pos_args
            kw_args += expanded_kw_args
            kw_values += expanded_kw_values

        else:
            non_expandable_args.append(v)


    for v in non_expandable_args:
        pos_args.append(
            ast.arg(
                arg=registry.generate_parameter_name(v.variable.name.value),
                annotation=recurse_type_annotation(v.type, registry),
            )
        )

    for v in kwarg_variables:
        kw_args.append(
            ast.arg(
                arg=registry.generate_parameter_name(v.variable.name.value),
                annotation=recurse_type_annotation(v.type, registry),
            )
        )
        kw_values.append(
            ast.Constant(
                value=parse_value_node(v.default_value) if v.default_value else None
            )
        )

    extra_kwargs = get_extra_kwargs_for_onode(definition, plugin_config)

    for kwarg in extra_kwargs:
        registry.register_import(kwarg.type)

        annotation = ast.Name(
            id=kwarg.type.split(".")[-1],
            ctx=ast.Load(),
        )

        if kwarg.default is None:
            # if we set the default to None, we need to make the annotation optional
            # complies with PEP 484
            annotation = ast.Subscript(
                value=ast.Name(id="Optional", ctx=ast.Load()),
                slice=annotation,
            )

        kw_args.append(
            ast.arg(
                arg=kwarg.key,
                annotation=annotation,
            )
        )
        kw_values.append(ast.Constant(value=kwarg.default))

    return ast.arguments(
        args=pos_args + kw_args,
        posonlyargs=[],
        kwonlyargs=[],
        kw_defaults=[],
        defaults=kw_values,
    )



def input_type_to_dict(input_type: GraphQLInputObjectType, registry: ClassRegistry):

    keys = []
    values = []

    for value_key, value in input_type.fields.items():
        field_name = registry.generate_node_name(value_key)

        keys.append(ast.Constant(value=value_key))
        values.append(ast.Name(id=field_name, ctx=ast.Load()))



    return ast.Dict(
        keys=keys,
        values=values,
    )







def generate_variable_dict(
    o: OperationDefinitionNode,
    plugin_config: FuncsPluginConfig,
    registry: ClassRegistry,
    client_schema: GraphQLSchema,
):
    keys = []
    values = []

    for v in o.variable_definitions:
        if v.variable.name.value in plugin_config.expand_input_types:
            input_type = v.type

            type = client_schema.type_map[input_type.type.name.value]



            keys.append(ast.Constant(value=v.variable.name.value))
            values.append(input_type_to_dict(type, registry))
        else:
            keys.append(ast.Constant(value=v.variable.name.value))
            values.append(
                ast.Name(
                    id=registry.generate_parameter_name(v.variable.name.value),
                    ctx=ast.Load(),
                )
            )

    return ast.Dict(keys=keys, values=values)


def generate_document_arg(o: OperationDefinitionNode, registry: ClassRegistry):
    return ast.Name(id=get_operation_class_name(o, registry), ctx=ast.Load())


def get_operation_class_name(
    o: OperationDefinitionNode, registry: ClassRegistry
) -> str:
    """Generates the name of the Operation Class for the given OperationDefinitionNode



    Args:
        o (OperationDefinitionNode): The graphql o node
        registry (ClassRegistry): The registry (used to get the operation class name)

    Raises:
        Exception: If the operation type is not supported

    Returns:
        str: _description_
    """

    if o.operation == OperationType.QUERY:
        return registry.style_query_class(o.name.value)
    if o.operation == OperationType.MUTATION:
        return registry.style_mutation_class(o.name.value)
    if o.operation == OperationType.SUBSCRIPTION:
        return registry.style_subscription_class(o.name.value)

    raise Exception("Incorrect Operation Type ")  # pragma: no cover


def get_return_type_annotation(
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    registry: ClassRegistry,
    collapse: bool = True,
) -> ast.AST:
    """Gets the return type annotation for the given operation definition node

    Ulized an autocollapse feature to collapse the return type annotation if it is a single fragment,
    to not generate unnecessary classes

    """

    o_name = get_operation_class_name(o, registry)
    root = get_operation_root_type(client_schema, o)


    if collapse is True:


        collapsable_field = o.selection_set.selections[0]

        sub_nodes = non_typename_fields(collapsable_field)
        field_definition = get_field_def(client_schema, root, collapsable_field)

        if len(sub_nodes) == 0:  # pragma: no cover
            return recurse_outputtype_annotation(field_definition.type, registry)
        

        if (
            len(sub_nodes) == 1
        ):  # Dealing with one Element
            collapsable_fragment_field = sub_nodes[0]
            if isinstance(
                collapsable_fragment_field, FragmentSpreadNode
            ):  # Dealing with a on element fragment
                return recurse_outputtype_annotation(
                    field_definition.type,
                    registry,
                    overwrite_final=registry.reference_fragment(
                        collapsable_fragment_field.name.value, "", allow_forward=False
                    ).id,
                )

        field_name = (
            collapsable_field.name.value
            if not collapsable_field.alias
            else collapsable_field.alias.value
        )

        return recurse_outputtype_annotation(
            field_definition.type,
            registry,
            overwrite_final=f"{o_name}{field_name.capitalize()}",
        )

    return ast.Name(
        id=o_name,
        ctx=ast.Load(),
    )


def get_return_type_string(
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    registry: ClassRegistry,
    collapse=True,
) -> Tuple[str, bool]:
    o_name = get_operation_class_name(o, registry)

    root = get_operation_root_type(client_schema, o)

    if collapse is True:
        
        potential_return_field = o.selection_set.selections[0]

        sub_nodes = non_typename_fields(potential_return_field)
        potential_return_type = get_field_def(
            client_schema, root, potential_return_field
        )

        if (
            len(sub_nodes) == 0
        ):  # Dealing with a scalar type  # pragma: no cover
            return recurse_outputtype_label(potential_return_type.type, registry)

        if (
            len(sub_nodes)  == 1
        ):  # Dealing with one Element
            collapsable_fragment_field = sub_nodes[0]

            if isinstance(
                collapsable_fragment_field, FragmentSpreadNode
            ):  # Dealing with a on element fragment
                return recurse_outputtype_label(
                    potential_return_type.type,
                    registry,
                    overwrite_final=registry.reference_fragment(
                        collapsable_fragment_field.name.value, "", allow_forward=False
                    ).id,
                )

        return recurse_outputtype_label(
            potential_return_type.type,
            registry,
            overwrite_final=f"{o_name}{potential_return_field.name.value.capitalize()}",
        )

    else:
        return o_name
    

def estimate_variable_name(o: OperationDefinitionNode, client_schema: GraphQLSchema, root: GraphQLObjectType, description_map: dict):

    if o.selection_set is None:
        return
    for field in o.selection_set.selections:
        if isinstance(field, FieldNode):


            for arg in field.arguments:
                the_field = get_field_def(client_schema, root, field)
                if isinstance(arg.value, VariableNode):
                    if arg.value.name.value in the_field.args:
                        mapped_description = the_field.args[arg.value.name.value].description

                        if mapped_description:
                            if arg.value.name.value in description_map:
                                description_map[arg.value.name.value] = f"{description_map[arg.value.name.value]} AND {mapped_description}"
                            else:
                                description_map[arg.value.name.value] = mapped_description

            for sub in non_typename_fields(field):
                if isinstance(sub, FieldNode):
                    estimate_variable_name(sub, client_schema, get_field_def(client_schema, root, field), description_map)








def generate_query_doc(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FuncsPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    x = get_operation_root_type(client_schema, o)
    o.__annotations__

    get_operation_class_name(o, registry)

    return_type = get_return_type_string(o, client_schema, registry, collapse)

    header = f"{o.name.value} \n"

    operation_documentation = (
        inspect_operation_for_documentation(o)
        if plugin_config.extract_documentation
        else None
    )

    description_map = {}
    estimate_variable_name(o, client_schema, x, description_map)


    if not operation_documentation:
        op_descriptions_dict = {}

        for field in o.selection_set.selections:
            if isinstance(field, FieldNode):
                target = target_from_node(field)
                operation_type = get_field_def(client_schema, x, field)
                if operation_type.description:
                    op_descriptions_dict[target] = f"{operation_type.description}"

        if collapse and len(op_descriptions_dict) == 1:
            op_descriptions = [
                f"{v}" for k, v in op_descriptions_dict.items()
            ]

        else:
            op_descriptions = [
                f"{k}: {v}" for k, v in op_descriptions_dict.items()
            ]



        description = "\n".join([header] + op_descriptions)

    else:
        description = header + operation_documentation

    description += "\n\nArguments:\n"

    extra_args = get_extra_args_for_onode(definition, plugin_config)

    for arg in extra_args:
        description += f"    {arg.key} ({arg.type}): {arg.description}\n"

    for v in o.variable_definitions:
        if v.variable.name.value in plugin_config.expand_input_types:

            input_type = v.type

            type = client_schema.type_map[input_type.type.name.value]


            description += generate_input_type_descriptions(type, registry)

        else:

            field_description = description_map.get(v.variable.name.value, "No description")


            if isinstance(v.type, NonNullTypeNode) and not v.default_value:
                description += f"    {registry.generate_parameter_name(v.variable.name.value)} ({recurse_type_label(v.type, registry)}): {field_description}\n"

    for v in o.variable_definitions:

        field_description = description_map.get(v.variable.name.value, "No description")

        if not isinstance(v.type, NonNullTypeNode) or v.default_value:
            description += f"    {registry.generate_parameter_name(v.variable.name.value)} ({recurse_type_label(v.type, registry)}, optional): {field_description}. {'' if not v.default_value else  'Defaults to ' + str(v.default_value.value)}\n"

    extra_kwargs = get_extra_kwargs_for_onode(definition, plugin_config)
    for kwarg in extra_kwargs:
        description += (
            f"    {kwarg.key} ({kwarg.type}, optional): {kwarg.description}\n"
        )

    description += "\nReturns:\n"
    description += f"    {return_type}\n"

    return ast.Expr(value=ast.Constant(value=description))


def genereate_async_call(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FuncsPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    registry.register_import(definition.use)

    if not collapse:
        return ast.Return(
            value=ast.Await(
                value=ast.Call(
                    func=ast.Name(
                        id=definition.use.split(".")[-1],
                        ctx=ast.Load(),
                    ),
                    keywords=generate_passing_extra_kwargs_for_onode(
                        definition, plugin_config
                    ),
                    args=generate_passing_extra_args_for_onode(
                        definition, plugin_config
                    )
                    + [
                        generate_document_arg(o, registry),
                        generate_variable_dict(o, plugin_config, registry, client_schema),
                    ],
                )
            )
        )
    else:
        correct_attr = (
            o.selection_set.selections[0].alias.value
            if o.selection_set.selections[0].alias
            else o.selection_set.selections[0].name.value
        )

        return ast.Return(
            value=ast.Attribute(
                value=ast.Await(
                    value=ast.Call(
                        func=ast.Name(
                            id=definition.use.split(".")[-1],
                            ctx=ast.Load(),
                        ),
                        keywords=generate_passing_extra_kwargs_for_onode(
                            definition, plugin_config
                        ),
                        args=generate_passing_extra_args_for_onode(
                            definition, plugin_config
                        )
                        + [
                            generate_document_arg(o, registry),
                            generate_variable_dict(o, plugin_config, registry, client_schema),
                        ],
                    )
                ),
                attr=registry.generate_node_name(correct_attr),
                ctx=ast.Load(),
            )
        )


def genereate_sync_call(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FuncsPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    registry.register_import(definition.use)
    if not collapse:
        return ast.Return(
            value=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o, registry),
                    generate_variable_dict(o, plugin_config, registry, client_schema),
                ],
            )
        )
    else:
        correct_attr = (
            o.selection_set.selections[0].alias.value
            if o.selection_set.selections[0].alias
            else o.selection_set.selections[0].name.value
        )

        return ast.Return(
            value=ast.Attribute(
                value=ast.Call(
                    func=ast.Name(
                        id=definition.use.split(".")[-1],
                        ctx=ast.Load(),
                    ),
                    keywords=generate_passing_extra_kwargs_for_onode(
                        definition, plugin_config
                    ),
                    args=generate_passing_extra_args_for_onode(
                        definition, plugin_config
                    )
                    + [
                        generate_document_arg(o, registry),
                        generate_variable_dict(o, plugin_config, registry, client_schema),
                    ],
                ),
                attr=registry.generate_node_name(correct_attr),
                ctx=ast.Load(),
            )
        )


def genereate_async_iterator(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FuncsPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    registry.register_import(definition.use)
    if not collapse:
        return ast.AsyncFor(
            target=ast.Name(id="event", ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o, registry),
                    generate_variable_dict(o, plugin_config, registry, client_schema),
                ],
            ),
            body=[
                ast.Expr(value=ast.Yield(value=ast.Name(id="event", ctx=ast.Load()))),
            ],
            orelse=[],
        )
    else:
        correct_attr = (
            o.selection_set.selections[0].alias.value
            if o.selection_set.selections[0].alias
            else o.selection_set.selections[0].name.value
        )

        return ast.AsyncFor(
            target=ast.Name(id="event", ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o, registry),
                    generate_variable_dict(o, plugin_config, registry, client_schema),
                ],
            ),
            body=[
                ast.Expr(
                    value=ast.Yield(
                        value=ast.Attribute(
                            value=ast.Name(id="event", ctx=ast.Load()),
                            ctx=ast.Load(),
                            attr=registry.generate_node_name(correct_attr),
                        )
                    )
                ),
            ],
            orelse=[],
        )


def genereate_sync_iterator(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FuncsPluginConfig,
    registry: ClassRegistry,
    collapse=False,
):
    registry.register_import(definition.use)
    if not collapse:
        return ast.For(
            target=ast.Name(id="event", ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o, registry),
                    generate_variable_dict(o, plugin_config, registry, client_schema),
                ],
            ),
            body=[
                ast.Expr(value=ast.Yield(value=ast.Name(id="event", ctx=ast.Load()))),
            ],
            orelse=[],
        )
    else:
        correct_attr = (
            o.selection_set.selections[0].alias.value
            if o.selection_set.selections[0].alias
            else o.selection_set.selections[0].name.value
        )

        return ast.For(
            target=ast.Name(id="event", ctx=ast.Store()),
            iter=ast.Call(
                func=ast.Name(
                    id=definition.use.split(".")[-1],
                    ctx=ast.Load(),
                ),
                keywords=generate_passing_extra_kwargs_for_onode(
                    definition, plugin_config
                ),
                args=generate_passing_extra_args_for_onode(definition, plugin_config)
                + [
                    generate_document_arg(o, registry),
                    generate_variable_dict(o, plugin_config, registry, client_schema),
                ],
            ),
            body=[
                ast.Expr(
                    value=ast.Yield(
                        value=ast.Attribute(
                            value=ast.Name(id="event", ctx=ast.Load()),
                            ctx=ast.Load(),
                            attr=registry.generate_node_name(correct_attr),
                        )
                    )
                ),
            ],
            orelse=[],
        )


def is_collapsable(o: OperationDefinitionNode):
    assert o.selection_set is not None, "Operation needs to have at least a selection"
    return len(o.selection_set.selections) == 1


def generate_operation_func(
    definition: FunctionDefinition,
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FuncsPluginConfig,
    registry: ClassRegistry,
):
    tree = []

    collapse = plugin_config.collapse_lonely and is_collapsable(o)

    return_type = get_return_type_annotation(
        o, client_schema, registry, collapse=collapse
    )

    doc = generate_query_doc(
        definition, o, client_schema, config, plugin_config, registry, collapse
    )

    if definition.is_async:
        if o.operation == OperationType.SUBSCRIPTION:
            registry.register_import("typing.AsyncIterator")

        tree.append(
            ast.AsyncFunctionDef(
                name=generate_async_func_name(o, plugin_config, config, registry),
                args=generate_parameters(
                    definition,
                    o,
                    config,
                    plugin_config,
                    registry,
                    client_schema
                ),
                body=[
                    doc,
                    (
                        genereate_async_call(
                            definition,
                            o,
                            client_schema,
                            config,
                            plugin_config,
                            registry,
                            collapse,
                        )
                        if definition.type != OperationType.SUBSCRIPTION
                        else genereate_async_iterator(
                            definition,
                            o,
                            client_schema,
                            config,
                            plugin_config,
                            registry,
                            collapse,
                        )
                    ),
                ],
                decorator_list=[],
                returns=(
                    return_type
                    if definition.type != OperationType.SUBSCRIPTION
                    else ast.Subscript(
                        value=ast.Name(id="AsyncIterator", ctx=ast.Load()),
                        slice=return_type,
                    )
                ),
            )
        )

    if not definition.is_async:
        if o.operation == OperationType.SUBSCRIPTION:
            registry.register_import("typing.Iterator")

        tree.append(
            ast.FunctionDef(
                name=generate_sync_func_name(o, plugin_config, config, registry),
                args=generate_parameters(
                    definition,
                    o,
                    config,
                    plugin_config,
                    registry,
                    client_schema
                ),
                body=[
                    doc,
                    (
                        genereate_sync_call(
                            definition,
                            o,
                            client_schema,
                            config,
                            plugin_config,
                            registry,
                            collapse,
                        )
                        if definition.type != OperationType.SUBSCRIPTION
                        else genereate_sync_iterator(
                            definition,
                            o,
                            client_schema,
                            config,
                            plugin_config,
                            registry,
                            collapse,
                        )
                    ),
                ],
                decorator_list=[],
                returns=(
                    return_type
                    if definition.type != OperationType.SUBSCRIPTION
                    else ast.Subscript(
                        value=ast.Name(id="Iterator", ctx=ast.Load()), slice=return_type
                    )
                ),
            )
        )

    return tree


class FuncsPlugin(Plugin):
    """This plugin generates functions for each operation in the schema.

    Contratry to the `operations` plugin, this plugin generates real python function
    with type annotations and docstrings according to the operation definition.

    These functions the can be used to call a proxy function (specified through the config)
    to execute the operation.

    You can also specify a list of extra arguments and keyword arguments that will be passed to the proxy function.

    Please consult the examples for more information.

    Example:

    ```python

    async def aexecute(operation: Model, variables: Dict[str, Any], client = None):
        client = client # is the grahql client that can be passed as an extra argument (or retrieved from a contextvar)
        x = await client.aquery(
            operation.Meta.document, operation.Arguments(**variables).dict(by_alias=True)
        )# is the proxy function that will be called (u can validate the variables here)
        return operation(**x.data) # Serialize the result

    ```

    Subscriptions are supported and will map to an async iterator.


    """

    config: FuncsPluginConfig = Field(default_factory=FuncsPluginConfig)

    def generate_ast(
        self,
        client_schema: GraphQLSchema,
        config: GeneratorConfig,
        registry: ClassRegistry,
    ) -> List[ast.AST]:
        plugin_tree = []

        documents = parse_documents(
            client_schema, self.config.funcs_glob or config.documents
        )

        operations = [
            node
            for node in documents.definitions
            if isinstance(node, OperationDefinitionNode)
        ]

        for operation in operations:
            for definition in get_definitions_for_onode(operation, self.config):
                plugin_tree += generate_operation_func(
                    definition,
                    operation,
                    client_schema,
                    config,
                    self.config,
                    registry,
                )

        return plugin_tree



from watchfiles.filters import BaseFilter
from watchfiles import watch
import os


class GraphQLFilter(BaseFilter):  # pragma: no cover
    def __call__(self, change, path: str) -> bool:
        x = super().__call__(change, path)
        if not x:
            return False

        x = os.path.basename(path)
        fileending = x.split(".")[1]

        return fileending == "graphql"


def stream_changes(folder: str):  # pragma: no cover
    for changes in watch(".", watch_filter=GraphQLFilter(), debounce=2000, step=500):
        yield changes



from pydantic import Field
from turms.stylers.base import BaseStyler, StylerConfig
import re


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


class SnakeCaseStylerConfig(StylerConfig):
    type: str = "turms.stylers.snake_case.SnakeCaseStyler"


class SnakeCaseStyler(BaseStyler):
    """A styler that snake cased node and parameter names.  (e.g. camelCase -> camel_case)"""

    config: SnakeCaseStylerConfig = Field(default_factory=SnakeCaseStylerConfig)

    def style_node_name(self, name: str) -> str:
        return camel_to_snake(name)

    def style_parameter_name(self, name: str) -> str:
        return camel_to_snake(name)
